from __future__ import annotations

import asyncio
import functools
import math
import typing

import rich.console
import toolcli
import toolstr
import tooltime

from ctc import cli
from ctc import evm
from ctc import spec
from ctc.defi import dex_utils
from ctc.defi.metric_utils import ohlc_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_dex_chart_command,
        'help': 'display candlestick chart of DEX pool trades',
        'args': [
            {
                'name': 'pool',
                'help': 'Uniswap pool address',
            },
            {
                'name': '--timescale',
                'help': 'size of candlesticks, e.g. 1h, 1d, or 1w',
            },
            {
                'name': '--invert',
                'action': 'store_true',
                'help': 'use inverse of price',
            },
            {
                'name': '--no-volume',
                'action': 'store_true',
                'help': 'hide volume data',
            },
        ],
        'examples': [
            '0xf4ad61db72f114be877e87d62dc5e7bd52df4d9b',
            '0xf4ad61db72f114be877e87d62dc5e7bd52df4d9b --invert',
        ],
    }


async def async_dex_chart_command(
    *,
    pool: spec.Address,
    invert: bool,
    timescale: str,
    no_volume: bool,
) -> None:

    dex = await dex_utils.async_get_dex_class(pool=pool)
    asset_symbols_task = asyncio.create_task(
        dex.async_get_pool_asset_symbols(pool=pool)
    )

    columns = toolcli.get_n_terminal_cols()
    n_candles = math.floor((columns - 10) / 2)
    if timescale is None:
        candle_timescale = '1d'
    else:
        candle_timescale = timescale
    candle_seconds = tooltime.timelength_to_seconds(candle_timescale)
    window_seconds = candle_seconds * n_candles

    window_end = tooltime.create_timestamp_seconds()
    window_start = (
        math.floor((window_end - window_seconds) / candle_seconds)
        * candle_seconds
    )
    start_block = await evm.async_get_block_of_timestamp(window_start) - 1
    end_block = await evm.async_get_latest_block_number()

    # get data
    swaps = await dex_utils.async_get_pool_trades(
        pool,
        start_block=start_block,
        end_block=end_block,
        normalize=True,
        include_volumes=True,
        include_prices=True,
    )

    # compute candlesticks
    prices = swaps['price__0__per__1'].to_numpy()
    x_volumes = swaps['volume__0'].to_numpy()
    if invert:
        prices = 1 / prices
    block_numbers = swaps['block_number'].to_list()
    block_timestamps = await evm.async_get_block_timestamps(
        block_numbers,
        context={'provider': {'chunk_size': 1}},
    )
    ohlc = ohlc_utils.compute_ohlc(
        values=list(prices),
        indices=block_timestamps,
        bin_size=candle_seconds,
        volumes=list(x_volumes),
    )
    ohlc = ohlc[-n_candles:]

    min_price = typing.cast(typing.Union[int, float], min(prices))
    max_price = typing.cast(typing.Union[int, float], max(prices))
    min_time = typing.cast(typing.Union[int, float], ohlc['bin'][0])
    max_time = typing.cast(typing.Union[int, float], ohlc['bin'][-1] + candle_seconds)
    render_grid = toolstr.create_grid(
        n_rows=20,
        n_columns=n_candles * 2,
        xmin=min_time - 0.05 * (max_time - min_time),
        xmax=max_time + 0.05 * (max_time - min_time),
        ymin=min_price - 0.05 * (max_price - min_price),
        ymax=max_price + 0.05 * (max_price - min_price),
    )
    sample_grid = toolstr.create_grid(sample_mode='quadrants', **render_grid)
    result = toolstr.raster_candlesticks(ohlc.rows(), sample_grid, render_grid)
    raster = result['raster']
    color_grid = result['color_grid']

    as_str = toolstr.render_supergrid(
        raster,
        char_dict='quadrants',
        color_grid=color_grid,
        color_map=toolstr.candlestick_color_map,
    )

    console = rich.console.Console(theme=rich.theme.Theme(inherit=False))

    styles = cli.get_cli_styles()
    tick_label_style = 'bold'
    chrome_style = '#888888'

    y_axis = toolstr.render_y_axis(
        grid=render_grid,
        tick_label_style=tick_label_style,
        chrome_style=chrome_style,
    )
    y_axis_width = rich.text.Text.from_markup(y_axis.split('\n')[0]).cell_len
    graph = toolstr.concatenate_blocks([y_axis, as_str])

    formatter = functools.partial(
        toolstr.format_timestamp,
        representation='TimestampDate',
    )
    x_axis = toolstr.render_x_axis(
        grid=render_grid,
        formatter=formatter,
        tick_label_style=tick_label_style,
        chrome_style=chrome_style,
    )
    x_axis = toolstr.indent_block(x_axis, indent=y_axis_width)

    # compute volume
    if not no_volume:
        ymax = float(ohlc['volume'].max()) * 1.1  # type: ignore
        volume_render_grid = toolstr.create_grid(
            n_rows=5,
            n_columns=n_candles * 2,
            xmin=render_grid['xmin'],
            xmax=render_grid['xmax'],
            ymin=0 - ymax / 9,
            ymax=ymax,
        )
        volume_sample_grid = toolstr.create_grid(
            sample_mode='quadrants',
            **volume_render_grid,
        )
        volume_raster = toolstr.raster_bar_chart(
            values=ohlc['volume'].to_list(),
            grid=volume_sample_grid,
            bar_width=1,
            bar_gap=3,
            start_gap=1,
        )
        volume_y_axis = toolstr.render_y_axis(
            grid=volume_render_grid,
            n_ticks=1,
            tick_label_style=tick_label_style,
            chrome_style=chrome_style,
        )

    # wait for metadata
    asset_symbols = await asset_symbols_task

    if len(asset_symbols) != 2:
        raise Exception('can only display candlesticks for pools with 2 assets')

    # print output
    token0 = asset_symbols[0]
    token1 = asset_symbols[1]
    toolstr.print_text_box(
        token0 + '-' + token1 + ' ' + dex.get_dex_name() + ' Pool',
        style=styles['title'],
    )
    cli.print_bullet(key='pool address', value=pool)
    cli.print_bullet(key='each candle', value=candle_timescale)
    cli.print_bullet(key='n_candles', value=n_candles)
    if invert:
        cli.print_bullet(
            key='price units', value=str(token1) + ' per ' + str(token0)
        )
    else:
        cli.print_bullet(
            key='price units', value=str(token0) + ' per ' + str(token1)
        )
    if not no_volume:
        cli.print_bullet(key='volume units', value=token0)
    print()
    console.print(graph)

    if not no_volume:
        volume_bars_str = toolstr.render_supergrid(
            volume_raster, char_dict='quadrants'
        )
        volume_graph = toolstr.concatenate_blocks(
            [volume_y_axis, volume_bars_str]
        )
        toolstr.print(volume_graph)

    # print x axis
    console.print(x_axis)



from __future__ import annotations

import typing

from ctc import evm
from ctc import rpc
from ctc import spec

from . import pool_metadata
from . import pool_parameters
from . import curve_spec


async def async_get_virtual_price(
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=curve_spec.pool_function_abis['get_virtual_price'],
        function_parameters=[],
        context=context,
        block_number=block,
    )

    if not isinstance(result, int):
        raise Exception('invalid rpc result')

    return result


async def async_get_lp_withdrawal(
    pool: spec.Address,
    amount_lp: int,
    *,
    token_withdrawn: spec.Address,
    context: spec.Context = None,
) -> int:

    token_index = await pool_metadata.async_get_token_index(
        pool=pool,
        token=token_withdrawn,
        context=context,
    )

    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=curve_spec.pool_function_abis['calc_withdraw_one_coin'],
        function_parameters=[amount_lp, token_index],
        context=context,
    )

    if not isinstance(result, int):
        raise Exception('invalid rpc result')

    return result


async def async_get_pool_state(
    pool: spec.Address,
    *,
    n_tokens: int | None = None,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
    normalize: bool = True,
) -> dict[str, typing.Any]:

    total_supply = await evm.async_get_erc20_total_supply(
        token=pool,
        block=block,
        context=context,
        normalize=normalize,
    )
    token_addresses = await pool_metadata.async_get_pool_tokens(
        pool,
        n_tokens=n_tokens,
        context=context,
    )
    token_balances = await evm.async_get_erc20s_balances(
        wallet=pool,
        tokens=token_addresses,
        block=block,
        context=context,
        normalize=normalize,
    )
    A = await pool_parameters.async_get_pool_A(
        pool=pool,
        block=block,
        context=context,
    )

    return {
        'lp_total_supply': total_supply,
        'token_balances': token_balances,
        'A': A,
    }


async def async_get_trade(
    pool: spec.Address,
    *,
    token_in: typing.Union[int, spec.Address],
    token_out: typing.Union[int, spec.Address],
    amount_in: typing.Union[int, float],
    input_normalized: bool = True,
    normalize_output: bool = True,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> typing.Union[int, float]:

    # get metadata
    metadata = await pool_metadata.async_get_pool_metadata(
        pool=pool,
        context=context,
    )
    in_index = await pool_metadata.async_get_token_index(
        pool=pool,
        token=token_in,
        metadata=metadata,
        context=context,
    )
    out_index = await pool_metadata.async_get_token_index(
        pool=pool,
        token=token_out,
        metadata=metadata,
        context=context,
    )

    # denormalize input
    if input_normalized:
        amount_in *= 10 ** metadata['token_decimals'][in_index]

    # query data
    amount_out = await rpc.async_eth_call(
        to_address=pool,
        function_abi=curve_spec.pool_function_abis['get_dy'],
        function_parameters=[
            in_index,
            out_index,
            amount_in,
        ],
        block_number=block,
        context=context,
    )

    # normalize output
    if normalize_output:
        amount_out = amount_out / (10 ** metadata['token_decimals'][out_index])

    if not isinstance(amount_out, (int, float)):
        raise Exception('invalid rpc result')

    return amount_out


# async def async_get_swaps():
#     """
#     Events
#     - TokenExchange 0x8b3e96f2b889fa771c53c981b40daf005f63f637f1869f707052d15a3dd97140
#     - TokenExchangeUnderlying 0xd013ca23e77a65003c2c659c5442c00c805371b7fc1ebd4c206c41d1536bd90b
#     """
#     pass




"""
these functions are not imported by default
- end users should never need to use these functions
"""

from __future__ import annotations

import os
import typing

import ctc
from ctc import spec

if typing.TYPE_CHECKING:
    import toolcli
    import toolsql
    import toolstr


def get_default_config(use_env_variables: bool = True) -> spec.Config:

    data_dir = get_default_data_dir()

    default_config: spec.Config = {
        'config_spec_version': ctc.__version__,
        'data_dir': data_dir,
        #
        # networks
        'default_network': 1,
        'networks': get_default_networks_metadata(),
        #
        # providers
        'providers': {},
        'default_providers': {},
        #
        # db
        'db_configs': get_default_db_configs(data_dir=data_dir),
        #
        # cache
        'context_cache_rules': [
            {
                'backend': 'main',
                'read': True,
                'write': True,
            },
        ],
        #
        # logging
        'log_rpc_calls': get_default_log_rpc_calls(),
        'log_sql_queries': get_default_log_sql_queries(),
        #
        # cli
        'cli_color_theme': get_default_cli_color_theme(),
        'cli_chart_charset': get_default_cli_chart_charset(),
    }
    # add in ETH_RPC_URL provider for no config mode
    if use_env_variables:
        add_env_var_rpc_provider(default_config)

    return default_config


def add_env_var_rpc_provider(default_config: spec.Config) -> None:
    eth_rpc_url = os.environ.get('ETH_RPC_URL')
    if eth_rpc_url is not None:
        print('using RPC provider stored in ETH_RPC_URL')
        provider_name = 'ETH_RPC_URL'

        # get chain id either 1) from env or 2) by querying node
        chain_id: int | str | None = os.environ.get('ETH_RPC_CHAIN_ID')
        if chain_id not in [None, '']:
            try:
                chain_id = int(chain_id)  # type: ignore
            except Exception:
                pass
        if chain_id is None:
            try:
                from ctc.rpc import rpc_provider

                chain_id = rpc_provider._sync_get_chain_id(eth_rpc_url)
            except Exception:
                print(
                    '[WARNING] not using value in ETH_RPC_URL because could not determine its chain_id (value = '
                    + str(eth_rpc_url)
                    + ')'
                )
                return
        if chain_id not in default_config['networks']:
            raise Exception(
                'cannot use provider in ETH_RPC_URL because it uses unknown chain_id = '
                + str(chain_id)
            )

        if not isinstance(chain_id, int) and chain_id is not None:
            raise Exception('could not determine proper value of chain_id')

        default_config['default_network'] = chain_id
        default_config['providers'][provider_name] = {  # type: ignore
            'name': provider_name,
            'network': chain_id,
            'protocol': 'http',
            'url': eth_rpc_url,
            'session_kwargs': {},
            'chunk_size': None,
            'convert_reverts_to_none': False,
            'disable_batch_requests': False,
        }
        default_config['default_providers'][chain_id] = provider_name  # type: ignore


def get_default_data_dir() -> str:
    return os.path.abspath(os.path.join(os.path.expanduser('~'), 'ctc_data'))


#
# # networks
#


def get_default_network_names_by_chain_id() -> typing.Mapping[int, str]:
    return {
        1: 'ethereum',
        3: 'ropsten',
        4: 'rinkeby',
        5: 'goerli',
        10: 'optimism',
        42: 'kovan',
        56: 'bnb',
        69: 'optimism_kovan',
        97: 'bnb_testnet',
        100: 'gnosis',
        128: 'heco',
        137: 'polygon',
        250: 'fantom',
        420: 'optimism_goerli',
        1088: 'metis',
        1284: 'moonbeam',
        1285: 'moonriver',
        4002: 'fantom_testnet',
        8217: 'klaytn',
        42161: 'arbitrum',
        43114: 'avalanche',
        43113: 'avalanche_fuji',
        80001: 'polygon_mumbai',
        421611: 'arbitrum_rinkeby',
        1666600000: 'harmony',
        1666700000: 'harmony_testnet',
    }


def get_default_chain_ids_by_network_name() -> typing.Mapping[str, int]:
    return {
        name: chain_id
        for chain_id, name in get_default_network_names_by_chain_id().items()
    }


def get_default_block_explorers() -> typing.Mapping[str, str]:
    return {
        'ethereum': 'etherscan.io',
        'ropsten': 'ropsten.etherscan.io',
        'rinkeby': 'rinkeby.etherscan.io',
        'goerli': 'goerli.etherscan.io',
        'optimism': 'optimistic.etherscan.io',
        'kovan': 'kovan.etherscan.io',
        'bnb': 'bscscan.com',
        'optimism_kovan': 'kovan-optimistic.etherscan.io',
        'bnb_testnet': 'testnet.bscscan.com',
        'gnosis': 'blockscout.com',
        'heco': 'hecoinfo.com',
        'polygon': 'polygonscan.com',
        'fantom': 'ftmscan.com',
        'optimism_goerli': 'blockscout.com/optimism/goerli',
        'metis': 'andromeda-explorer.metis.io',
        'moonbeam': 'moonbeam.moonscan.io',
        'moonriver': 'moonriver.moonscan.io',
        'fantom_testnet': 'testnet.ftmscan.com',
        'klaytn': 'scope.klaytn.com',
        'arbitrum': 'arbiscan.io',
        'avalanche': 'snowtrace.io',
        'avalanche_fuji': 'testnet.snowtrace.io',
        'polygon_mumbai': 'mumbai.polygonscan.com',
        'arbitrum_rinkeby': 'testnet.arbiscan.io',
        'harmony': 'explorer.harmony.one',
        'harmony_testnet': 'explorer.testnet.harmony.one',
    }


def get_default_networks_metadata() -> typing.Mapping[
    spec.ChainId, spec.NetworkMetadata
]:
    block_explorers = get_default_block_explorers()
    chain_ids_by_network_name = get_default_chain_ids_by_network_name()
    return {
        chain_id: {
            'name': network_name,
            'chain_id': chain_id,
            'block_explorer': block_explorers[network_name],
        }
        for network_name, chain_id in chain_ids_by_network_name.items()
    }


#
# # db
#


def get_default_db_config(*, data_dir: str) -> toolsql.DBConfig:
    return {
        'dbms': 'sqlite',
        'path': os.path.join(data_dir, 'dbs/ctc.db'),
    }


def get_default_db_configs(
    *,
    data_dir: str,
) -> typing.Mapping[str, toolsql.DBConfig]:
    return {'main': get_default_db_config(data_dir=data_dir)}


#
# # logging
#


def get_default_log_rpc_calls() -> bool:
    return False


def get_default_log_sql_queries() -> bool:
    return False


#
# # cli
#


def get_default_cli_color_theme() -> toolcli.StyleTheme:
    return {
        'title': 'bold #ce93f9',
        'metavar': '#8be9fd',
        'description': '#b9f29f',
        'content': '#f1fa8c',
        'option': '#64aaaa',
        'comment': '#6272a4',
    }


def get_default_cli_chart_charset() -> toolstr.SampleMode:
    return 'braille'




from __future__ import annotations

import os
import typing

import toolcli
import toolstr

from ctc import cli
import ctc.config


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': config_command,
        'help': 'print current config information',
        'args': [
            {
                'name': '--reveal',
                'action': 'store_true',
                'help': 'show sensitive information in config',
            },
            {
                'name': '--json',
                'help': 'output config as json',
                'dest': 'as_json',
                'action': 'store_true',
            },
            {
                'name': '--no-color',
                'action': 'store_true',
                'help': 'do not use color',
            },
        ],
        'examples': [
            '',
            '--reveal',
            '--json',
        ],
    }


def config_command(*, reveal: bool, as_json: bool, no_color: bool) -> None:
    env_var = ctc.config.config_path_env_var

    if as_json:
        config: typing.Mapping[str, typing.Any] = ctc.config.get_config()

        if no_color:
            import json

            print(json.dumps(config, indent=4, sort_keys=True))
        else:
            import rich

            rich.print(config)

    else:
        # get styles
        if no_color:
            title_style = None
            chrome_style = None
            content_style = None
            description_style = None
            path_style = None
            key_style = None

        else:
            styles = cli.get_cli_styles()
            title_style = styles['title']
            chrome_style = styles['comment']
            key_style = styles['option']
            content_style = styles['content']
            description_style = styles['description']
            path_style = styles['metavar']

            path_style = path_style + ' bold'

        toolstr.print_text_box(
            'Config Summary', style=title_style, text_style=path_style
        )

        # print config path
        env_var_value = os.environ.get(env_var)
        config_path = ctc.config.get_config_path(raise_if_dne=False)

        styled_path = toolstr.add_style(config_path, path_style)
        if env_var_value in (None, ''):
            toolstr.print(
                toolstr.add_style('using default config path: ', key_style)
                + styled_path
            )
        else:
            toolstr.print(
                toolstr.add_style(
                    'using config path in CTC_CONFIG_PATH:', key_style
                ),
                styled_path,
            )

        print()
        print()
        toolstr.print_header(
            'Environment Variables', style=title_style, text_style=path_style
        )
        for env_var in [
            ctc.config.config_path_env_var,
            ctc.config.provider_env_var,
            ctc.config.network_env_var,
            ctc.config.cache_env_var,
        ]:
            toolstr.print_bullet(
                key=env_var, value=os.environ.get(env_var), styles=styles
            )
        toolstr.print('(these are all optional)', style=chrome_style)

        print()
        print()
        toolstr.print_header(
            'Config Values', style=title_style, text_style=path_style
        )
        config = typing.cast(
            typing.Mapping[str, typing.Any], ctc.config.get_config()
        )

        for key in sorted(config.keys()):
            if key == 'cli_color_theme':
                toolstr.print(
                    toolstr.add_style('-', chrome_style),
                    toolstr.add_style(key, key_style)
                    + toolstr.add_style(':', chrome_style),
                )
                for subkey, subvalue in config[key].items():
                    toolstr.print(
                        '    ' + toolstr.add_style('-', chrome_style),
                        toolstr.add_style(subkey, key_style)
                        + toolstr.add_style(':', chrome_style),
                        toolstr.add_style(subvalue, subvalue),
                    )

            elif key == 'networks':
                print()
                toolstr.print(
                    toolstr.add_style('-', chrome_style),
                    toolstr.add_style(key, key_style)
                    + toolstr.add_style(':', chrome_style),
                )
                rows = []
                for chain_id, network_metadata in config[key].items():
                    row = [
                        network_metadata['name'],
                        str(network_metadata['chain_id']),
                        network_metadata['block_explorer'],
                    ]
                    rows.append(row)
                labels = ['name', 'chain id', 'block_explorer']
                toolstr.print_table(
                    rows,
                    labels=labels,
                    indent=4,
                    label_style=title_style,
                    column_styles={
                        'name': description_style,
                        'chain id': content_style,
                        'block_explorer': path_style,
                    },
                    border=chrome_style,
                )
                print()

            elif key == 'providers':
                toolstr.print(
                    toolstr.add_style('-', chrome_style),
                    toolstr.add_style(key, key_style)
                    + toolstr.add_style(':', chrome_style),
                )
                if len(config[key]) == 0:
                    continue
                labels = ['name', 'network', 'chain id', 'url']
                rows = []
                for provider in config[key].values():
                    row = []
                    for label in labels:
                        if label == 'url' and not reveal:
                            cell = '*' * 8
                        elif label == 'network':
                            cell = config['networks'][provider['network']][
                                'name'
                            ]
                        elif label == 'chain id':
                            cell = config['networks'][provider['network']][
                                'chain_id'
                            ]
                        elif label == 'name':
                            cell = provider['name']
                        else:
                            cell = str(provider[label])
                        row.append(cell)
                    rows.append(row)
                toolstr.print_table(
                    rows,
                    labels=labels,
                    indent=4,
                    label_style=title_style,
                    border=chrome_style,
                    column_styles={
                        'name': description_style,
                        'network': description_style,
                        'url': path_style,
                        'chain id': content_style,
                    },
                    column_formats={
                        'chain id': {'commas': False},
                    },
                    sort_column=['chain id', 'name'],
                )

                if not reveal:
                    print()
                    toolstr.print(
                        toolstr.add_style('    (use ', chrome_style)
                        + toolstr.add_style('--reveal', key_style)
                        + toolstr.add_style(
                            ' to reveal sensitive provider information)',
                            chrome_style,
                        )
                    )

            elif isinstance(config[key], dict) and len(config[key]) > 0:
                toolstr.print(
                    toolstr.add_style('-', chrome_style),
                    toolstr.add_style(key, key_style)
                    + toolstr.add_style(':', chrome_style),
                )
                for subkey, subvalue in config[key].items():
                    if isinstance(subvalue, dict) and len(subvalue) > 0:
                        toolstr.print(
                            toolstr.add_style('    -', chrome_style),
                            toolstr.add_style(str(subkey), description_style)
                            + toolstr.add_style(':', chrome_style),
                        )
                        for subsubkey, subsubvalue in subvalue.items():
                            if key == 'db_configs' and subsubkey == 'path':
                                style = path_style
                            else:
                                style = description_style

                            toolstr.print(
                                toolstr.add_style('        -', chrome_style),
                                toolstr.add_style(
                                    str(subsubkey), description_style
                                )
                                + toolstr.add_style(':', chrome_style),
                                toolstr.add_style(str(subsubvalue), style),
                            )
                    else:
                        toolstr.print(
                            toolstr.add_style('    -', chrome_style),
                            toolstr.add_style(str(subkey), description_style)
                            + toolstr.add_style(':', chrome_style),
                            toolstr.add_style(str(subvalue), description_style),
                        )

            elif key == 'context_cache_rules':
                rules = config[key]
                rows = []
                for rule in rules:
                    row = []
                    row.append(rule.get('backend', ''))
                    row.append(rule.get('read', ''))
                    row.append(rule.get('write', ''))
                    filter = rule.get('filter')
                    if filter is None or len(filter) == 0:
                        scope = 'global'
                    else:
                        scope = ','.join(
                            (key + '=' + value) for key, value in filter.items()
                        )
                    row.append(scope)
                    rows.append(row)

                toolstr.print_bullet(
                    key='cache configuration', value='', key_style=key_style
                )
                labels = ['backend', 'read', 'write', 'scope']
                toolstr.print_table(
                    rows,
                    labels=labels,
                    indent=8,
                    label_style=title_style,
                    border=chrome_style,
                    column_styles={
                        'backend': path_style,
                        'read': description_style,
                        'write': description_style,
                        'scope': content_style,
                    },
                    add_row_index=True,
                )
                toolstr.print(
                    '(rules applied in order)',
                    style=chrome_style,
                    indent=8,
                )

            else:
                if key == 'data_dir':
                    style = path_style
                else:
                    style = description_style

                toolstr.print(
                    toolstr.add_style('-', chrome_style),
                    toolstr.add_style(key, key_style)
                    + toolstr.add_style(':', chrome_style),
                    toolstr.add_style(str(config[key]), style),
                )




from __future__ import annotations

import typing

from ctc import config
from ctc import rpc
from ctc import spec


feed_registry = {
    # ethereum
    1: '0x47fb2585d2c56fe188d0e6ec628a38b74fceeedf',
    # kovan
    42: '0xAa7F6f7f507457a1EE157fE97F6c7DB2BEec5cD0',
}

feed_registry_asset_addresses = {
    'BTC': '0xbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb',
    'ETH': '0xeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee',
    'USD': '0x0000000000000000000000000000000000000348',
    'GBP': '0x000000000000000000000000000000000000033a',
    'EUR': '0x00000000000000000000000000000000000003d2',
}

feed_registry_address_assets = {
    v: k for k, v in feed_registry_asset_addresses.items()
}

registry_function_abis: typing.Mapping[str, spec.FunctionABI] = {
    'getFeed': {
        'inputs': [
            {'internalType': 'address', 'name': 'base', 'type': 'address'},
            {'internalType': 'address', 'name': 'quote', 'type': 'address'},
        ],
        'name': 'getFeed',
        'outputs': [
            {
                'internalType': 'contract AggregatorV2V3Interface',
                'name': 'aggregator',
                'type': 'address',
            }
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'getPhaseRange': {
        'inputs': [
            {'internalType': 'address', 'name': 'base', 'type': 'address'},
            {'internalType': 'address', 'name': 'quote', 'type': 'address'},
            {'internalType': 'uint16', 'name': 'phaseId', 'type': 'uint16'},
        ],
        'name': 'getPhaseRange',
        'outputs': [
            {
                'internalType': 'uint80',
                'name': 'startingRoundId',
                'type': 'uint80',
            },
            {
                'internalType': 'uint80',
                'name': 'endingRoundId',
                'type': 'uint80',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'getPhase': {
        'inputs': [
            {'internalType': 'address', 'name': 'base', 'type': 'address'},
            {'internalType': 'address', 'name': 'quote', 'type': 'address'},
            {'internalType': 'uint16', 'name': 'phaseId', 'type': 'uint16'},
        ],
        'name': 'getPhase',
        'outputs': [
            {
                'components': [
                    {
                        'internalType': 'uint16',
                        'name': 'phaseId',
                        'type': 'uint16',
                    },
                    {
                        'internalType': 'uint80',
                        'name': 'startingAggregatorRoundId',
                        'type': 'uint80',
                    },
                    {
                        'internalType': 'uint80',
                        'name': 'endingAggregatorRoundId',
                        'type': 'uint80',
                    },
                ],
                'internalType': 'struct FeedRegistryInterface.Phase',
                'name': 'phase',
                'type': 'tuple',
            }
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
}


async def async_get_registry_feed(
    base: str,
    quote: str,
    *,
    context: spec.Context = None,
) -> spec.Address:

    network = config.get_context_chain_id(context)
    registry_address = feed_registry[network]

    result = await rpc.async_eth_call(
        to_address=registry_address,
        function_abi=registry_function_abis['getFeed'],
        function_parameters=[base, quote],
        context=context,
    )

    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result


async def async_get_phase_range(
    *,
    base: str,
    quote: str,
    phase: int,
    context: spec.Context = None,
) -> typing.Tuple[int, ...]:

    network = config.get_context_chain_id(context)
    registry_address = feed_registry[network]

    result = await rpc.async_eth_call(
        to_address=registry_address,
        # function_abi=registry_function_abis['getPhaseRange'],
        function_abi=registry_function_abis['getPhase'],
        function_parameters=[base, quote, phase],
        context=context,
    )

    if not isinstance(result, tuple) or not all(
        isinstance(item, int) for item in result
    ):
        raise Exception('invalid rpc result')
    return result




import os
import typing

if typing.TYPE_CHECKING or os.environ.get('BUILDING_SPHINX') == '1':
    from .abi_types import *
    from .address_types import *
    from .binary_types import *
    from .block_types import *
    from .config_types import *
    from .consensus_types import *
    from .context_types import *
    from .data_source_types import *
    from .db_types import *
    from .defi_types import *
    from .event_types import *
    from .external_types import *
    from .log_types import *
    from .network_types import *
    from .number_types import *
    from .rpc_types import *
    from .storage_types import *
    from .trace_types import *
    from .transaction_types import *




from __future__ import annotations

import typing
from ctc import spec
from . import consensus_requests


#
# # epochs
#


async def async_get_consensus_finalized_epoch(
    *,
    state_id: str = 'head',
    context: spec.Context = None,
) -> int:
    result: spec.ConsensusFinalityCheckpoints = (
        await consensus_requests.async_beacon_request(
            'beacon_finality_checkpoints',
            params=dict(state_id=state_id),
            context=context,
        )
    )
    return int(result['finalized']['epoch'])


async def async_get_consensus_justified_epoch(
    *,
    state_id: str = 'head',
    context: spec.Context = None,
) -> int:
    result: spec.ConsensusFinalityCheckpoints = (
        await consensus_requests.async_beacon_request(
            'beacon_finality_checkpoints',
            params=dict(state_id=state_id),
            context=context,
        )
    )
    return int(result['current_justified']['epoch'])


#
# # slots
#


async def async_get_consensus_current_slot(
    *,
    context: spec.Context = None,
) -> int:
    result: typing.Any = await consensus_requests.async_beacon_request(
        'beacon_syncing', context=context
    )
    return int(result['head_slot'])


#
# # validators
#


async def async_get_consensus_validators(
    *,
    state_id: str = 'head',
    context: spec.Context = None,
) -> typing.Sequence[spec.ConsensusValidator]:
    result: typing.Sequence[
        spec.ConsensusValidator
    ] = await consensus_requests.async_beacon_request(
        'beacon_validators',
        params=dict(state_id=state_id),
        context=context,
    )
    return result


async def async_get_consensus_validator(
    *,
    state_id: str = 'head',
    context: spec.Context = None,
) -> spec.ConsensusValidator:

    result: spec.ConsensusValidator = (
        await consensus_requests.async_beacon_request(
            'beacon_validator',
            params=dict(state_id=state_id),
            context=context,
        )
    )
    return result


async def async_get_consensus_validator_count(
    *,
    state_id: str = 'head',
    context: spec.Context = None,
) -> int:
    validators = await async_get_consensus_validators(
        state_id=state_id,
        context=context,
    )
    return len(validators)


#
# # peers
#


async def async_get_consensus_peers(
    *,
    context: spec.Context = None,
) -> typing.Sequence[spec.ConsensusPeer]:
    result: typing.Sequence[
        spec.ConsensusPeer
    ] = await consensus_requests.async_beacon_request(
        'node_peers',
        context=context,
    )
    return result


async def async_get_consensus_peer(
    peer_id: str,
    *,
    context: spec.Context = None,
) -> spec.ConsensusPeer:
    result: spec.ConsensusPeer = await consensus_requests.async_beacon_request(
        'node_peer',
        params=dict(peer_id=peer_id),
        context=context,
    )
    return result


async def async_get_consensus_peer_count(
    *,
    context: spec.Context = None,
) -> typing.Mapping[str, int]:
    result: typing.Mapping[
        str, str
    ] = await consensus_requests.async_beacon_request(
        'node_peer_count',
        context=context,
    )
    return {k: int(v) for k, v in result.items()}


#
# # setup
#


async def async_get_consensus_deposit_contract(
    *,
    context: spec.Context = None,
) -> spec.Address:
    result: spec.Address = await consensus_requests.async_beacon_request(
        'config_deposit_contract',
        context=context,
    )
    return result


#
# # node
#


async def async_get_consensus_client_version(
    *,
    context: spec.Context = None,
) -> str:
    result: str = await consensus_requests.async_beacon_request(
        'node_version',
        context=context,
    )
    return result


async def async_get_consensus_client_id(
    *,
    context: spec.Context = None,
) -> str:
    result: str = await consensus_requests.async_beacon_request(
        'node_identity',
        context=context,
    )
    return result




from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import tooltime

import matplotlib.pyplot as plt  # type: ignore
import toolstr

from ctc import evm
from ctc import spec


if typing.TYPE_CHECKING:
    T = typing.TypeVar('T')


async def async_xtick_block_dates(
    context: spec.Context = None,
) -> None:
    import tooltime

    start_block, end_block = plt.xlim()
    start_time, end_time = await evm.async_predict_block_timestamps(
        [round(start_block), round(end_block)], context=context
    )
    date_timestamps = tooltime.get_standard_intervals(
        interval_size='1d',
        start_time=start_time,
        end_time=end_time,
    )
    date_timestamps = [
        timestamp
        for timestamp in date_timestamps
        if timestamp >= start_time and timestamp <= end_time
    ]
    selected_dates = _select_interleaved(date_timestamps, 5)
    blocks_per_second = (end_block - start_block) / (end_time - start_time)
    locations = [
        start_block + blocks_per_second * (timestamp - start_time)
        for timestamp in selected_dates
    ]
    labels = [
        tooltime.convert_timestamp(timestamp, 'TimestampDate')
        for timestamp in selected_dates
    ]
    plt.xticks(locations, labels, ha='left', rotation=-25)


def _select_interleaved(
    items: typing.Sequence[T], number: int
) -> typing.Sequence[T]:
    import numpy as np

    if len(items) == 0:
        return []
    elif len(items) <= number:
        return items
    else:
        indices = np.linspace(0, len(items) - 1, number).round().astype(int)
        return [items[index] for index in indices]


async def async_block_timestamp_xticks(
    context: spec.Context = None,
    representation: tooltime.TimestampExtendedRepresentation | None = None,
    *,
    omit: str | None = 'year',
) -> None:
    """
    matplotlib.FuncFormatter not used, because it cannot use async
    """

    raw_blocks, labels = plt.xticks()
    blocks = [int(block) for block in raw_blocks]
    block_timestamps = await evm.async_predict_block_timestamps(
        blocks, context=context
    )

    if representation is None:
        if len(block_timestamps) < 2:
            representation = 'TimestampDate'
        else:
            interval = block_timestamps[1] - block_timestamps[0]
            if interval > 86400:
                representation = 'TimestampDate'
            else:
                representation = 'TimestampISOPretty'

    labels = [
        toolstr.format(
            timestamp,
            format_type='timestamp',
            representation=representation,
        )
        for timestamp in block_timestamps
    ]

    # need to decide cleaner locations when dealing with shorter timescales
    # probably need to aim for somewhere between 4 and 6 ticks
    # should see how many ticks result from
    # - day intervals
    # - 12 hour intervals
    # - 6 hour intervals
    # - 4 hour intervals
    # - 3 hour intervals
    # - 2 hour intervals
    # - 1 hour intervals
    # - 30 minute intervals
    # - 20 minute intervals
    # - 10 minute intervals
    # - 5 minute intervals
    # - 2 minute intervals
    # - 1 minute intervals
    # - 30 second intervals
    # - 20 second intervals
    # - 10 second intervals
    # go for whichever size brings you closest to 5 ticks

    if representation == 'TimestampISOPretty' and omit is not None:
        if omit == 'year':
            labels = [label[5:] for label in labels]

    plt.xticks(raw_blocks, labels)



from __future__ import annotations

import typing

from ctc import spec
from ctc.toolbox import nested_utils
from . import directory_metadata
from . import pool_metadata
from . import token_metadata
from . import token_state

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    class CTokenMetricSpec(TypedDict, total=False):
        tvl: bool
        tvb: bool
        supply_apy: bool
        borrow_apy: bool

    class CTokenMetrics(TypedDict, total=False):
        tvl: spec.Number
        tvb: spec.Number
        supply_apy: spec.Number | None
        borrow_apy: spec.Number | None

    class CTokenMetricsByBlock(TypedDict, total=False):
        tvl: typing.Sequence[spec.Number]
        tvb: typing.Sequence[spec.Number]
        supply_apy: typing.Sequence[spec.Number | None]
        borrow_apy: typing.Sequence[spec.Number | None]


async def async_get_token_multipool_history(
    token: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    metrics: CTokenMetricSpec | None = None,
    comptrollers: typing.Sequence[spec.Address] | None = None,
) -> dict[spec.Address, CTokenMetricsByBlock]:
    """get token metrics across multiple pools and blocks"""

    import asyncio

    if comptrollers is None:
        pools = await directory_metadata.async_get_all_pools()
        comptrollers = [pool[2] for pool in pools]

    coroutines = []
    for comptroller in comptrollers:
        coroutine = async_get_pool_token_history(
            comptroller, token, blocks=blocks, metrics=metrics
        )
        coroutines.append(coroutine)
    results = await asyncio.gather(*coroutines)

    return {
        comptroller: result
        for comptroller, result in zip(comptrollers, results)
        if result is not None
    }


async def async_get_pool_token_history(
    comptroller: spec.Address,
    token: spec.Address,
    *,
    blocks: typing.Sequence[spec.BlockNumberReference],
    metrics: CTokenMetricSpec | None = None,
) -> CTokenMetricsByBlock | None:
    """get metrics of a token in a pool across blocks"""
    import asyncio

    ctokens = await pool_metadata.async_get_pool_ctokens(comptroller)
    coroutines = []
    for ctoken in ctokens:
        coroutine = _async_get_token_history_if_match(
            ctoken, token, blocks=blocks, metrics=metrics
        )
        coroutines.append(coroutine)
    results = await asyncio.gather(*coroutines)

    matching_results = [result for result in results if result is not None]

    if len(matching_results) == 0:
        return None
    elif len(matching_results) == 1:
        return matching_results[0]  # type: ignore
    else:
        raise Exception('more than one ctoken found for underlying')


async def _async_get_token_history_if_match(
    ctoken: spec.Address,
    token: spec.Address,
    *,
    blocks: typing.Sequence[spec.BlockNumberReference],
    metrics: CTokenMetricSpec | None = None,
) -> CTokenMetricsByBlock | None:
    underlying = await token_metadata.async_get_ctoken_underlying(ctoken)
    if underlying != token.lower():
        return None
    else:
        return await async_get_ctoken_state_by_block(
            ctoken=ctoken,
            blocks=blocks,
            metrics=metrics,
            in_usd=False,
        )


async def async_get_ctoken_state_by_block(
    ctoken: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    metrics: CTokenMetricSpec | None = None,
    eth_price: spec.Number | None = None,
    in_usd: bool = True,
) -> CTokenMetricsByBlock:
    import asyncio

    coroutines = []
    for block in blocks:
        coroutine = async_get_ctoken_state(
            ctoken,
            block=block,
            metrics=metrics,
            eth_price=eth_price,
            in_usd=in_usd,
        )
        coroutines.append(coroutine)
    results = await asyncio.gather(*coroutines)
    return typing.cast(
        CTokenMetricsByBlock,
        nested_utils.list_of_dicts_to_dict_of_lists(results),
    )


async def async_get_ctoken_state(
    ctoken: spec.Address,
    *,
    block: spec.BlockNumberReference = 'latest',
    metrics: CTokenMetricSpec | None = None,
    eth_price: spec.Number | None = None,
    in_usd: bool = True,
) -> CTokenMetrics:
    import asyncio

    if metrics is None:
        metrics = {
            'tvl': True,
            'tvb': True,
            'supply_apy': True,
            'borrow_apy': True,
        }

    blocks_per_year = 2102400
    if metrics.get('tvl') or metrics.get('tvb'):
        tv_coroutine = token_state.async_get_ctoken_tvl_and_tvb(
            ctoken=ctoken,
            block=block,
            in_usd=in_usd,
            eth_price=eth_price,
        )
        tv_task = asyncio.create_task(tv_coroutine)
    if metrics.get('supply_apy'):
        coroutine = token_state.async_get_supply_apy(
            ctoken=ctoken,
            block=block,
            blocks_per_year=blocks_per_year,
            fill_empty=True,
            empty_token=0,
        )
        supply_task = asyncio.create_task(coroutine)
    if metrics.get('borrow_apy'):
        coroutine = token_state.async_get_borrow_apy(
            ctoken=ctoken,
            block=block,
            blocks_per_year=blocks_per_year,
            fill_empty=True,
            empty_token=0,
        )
        borrow_task = asyncio.create_task(coroutine)

    # gather output
    output: CTokenMetrics = {}
    if metrics.get('tvl') or metrics.get('tvb'):
        tv_result = await tv_task
        if metrics.get('tvl'):
            output['tvl'] = tv_result['tvl']
        if metrics.get('tvb'):
            output['tvb'] = tv_result['tvb']
    if metrics.get('supply_apy'):
        output['supply_apy'] = await supply_task
    if metrics.get('borrow_apy'):
        output['borrow_apy'] = await borrow_task

    return output




from __future__ import annotations

import toolsql

fuse_pools_schema: toolsql.DBSchemaShorthand = {
    'tables': {
        'fuse_pools': {
            'columns': [
                {'name': 'index', 'type': 'Integer', 'primary': True},
                {'name': 'comptroller', 'type': 'Text', 'index': True},
            ],
        },
        'fuse_ftokens': {
            'columns': [
                {'name': 'address', 'type': 'Text', 'primary': True},
                {'name': 'underyling', 'type': 'Text', 'index': True},
                {'name': 'comptroller', 'type': 'Text', 'index': True},
            ],
        },
    },
}



from __future__ import annotations

import asyncio
import typing

import toolcli
import toolstr

from ctc import cli
from ctc import evm
from ctc import rpc
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_call_all_command,
        'help': 'display outputs of all contract\'s read-only functions'
        + '\n\ncurrently only displays output of zero-parameter functions',
        'args': [
            {
                'name': 'contract_address',
                'help': 'address of contract to call functions of',
            },
            {
                'name': '--max-results',
                'default': 10,
                'type': int,
                'help': 'for functions that output lists, limit display length',
            },
            {
                'name': '--block',
                'help': 'block number for call',
            },
        ],
        'examples': [
            '0x6c3f90f043a72fa612cbac8115ee7e52bde6e490',
            '0x240315db938d44bb124ae619f5fd0269a02d1271 --max-results 5',
        ],
    }


async def async_call_all_command(
    *,
    contract_address: spec.Address,
    max_results: int,
    block: str,
) -> None:

    if block is not None:
        block_number = await evm.async_block_number_to_int(block)
    else:
        block_number = 'latest'

    contract_abi = await evm.async_get_contract_abi(contract_address)

    function_abis = []
    coroutines = []
    for item in contract_abi:
        if item.get('type') == 'function':
            if typing.TYPE_CHECKING:
                function_abi = typing.cast(spec.FunctionABI, item)
            else:
                function_abi = item
            if evm.is_function_read_only(function_abi):
                inputs = function_abi.get('inputs')
                if inputs is not None and len(inputs) == 0:
                    function_abis.append(function_abi)
                    coroutine = rpc.async_eth_call(
                        to_address=contract_address,
                        function_abi=function_abi,
                        context=dict(provider={'convert_reverts_to_none': True}),
                        block_number=block_number,
                    )
                    coroutines.append(coroutine)

    results = await asyncio.gather(*coroutines)

    styles = cli.get_cli_styles()

    multiline = False
    rows = []
    for r in range(len(results)):
        function_abi = function_abis[r]
        row = [function_abi['name']]
        result = results[r]

        # modify specific result values
        if result is None:
            result = toolstr.add_style('\[REVERT]', styles['title'])
        if isinstance(result, int):
            result = str(result)

        # add result to row
        if isinstance(result, (tuple, list)):
            import json

            footnote = 'length = ' + str(len(result))
            if len(result) > max_results:
                footnote += ', clipped to ' + str(max_results)
                result = result[:max_results]
                clipped = True
            else:
                clipped = False
            try:
                str_result = json.dumps(result, sort_keys=True, indent=2)
            except TypeError:
                str_result = str(result)
            if clipped:
                str_result = str_result[: str_result.rindex('\n')]
                str_result = str_result + ',\n...'

            footnote = toolstr.add_style(
                '(' + footnote + ')', styles['comment']
            )
            str_result += '\n\n' + footnote
            row.append(str_result)
            multiline = True
        else:
            row.append(result)

        rows.append(row)

    labels = [
        'function',
        'output',
    ]

    toolstr.print_text_box(
        'Function Outputs for ' + contract_address, style=styles['title']
    )
    print()
    if multiline:
        toolstr.print_multiline_table(
            rows,
            labels=labels,
            indent=4,
            border=styles['comment'],
            label_style=styles['title'],
            column_styles={
                'function': styles['option'],
                'output': styles['description'],
            },
            column_justify={'output': 'raw'},
        )
    else:
        toolstr.print_table(
            rows,
            labels=labels,
            indent=4,
            border=styles['comment'],
            label_style=styles['title'],
            column_styles={
                'function': styles['option'],
                'output': styles['description'],
            },
        )



from __future__ import annotations

import typing

import toolcli
import toolstr

from ctc import cli
from ctc import evm
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_proxy_command,
        'help': 'print proxy information about contract',
        'args': [
            {'name': 'contract_address', 'help': 'address to get proxy of'},
            {
                'name': ['--verbose', '-v'],
                'help': 'print additional information',
                'action': 'store_true',
            },
            {'name': '--block', 'help': 'block to query', 'default': 'latest'},
        ],
        'examples': [
            '0xee6a57ec80ea46401049e92587e52f5ec1c24785',  # EIP1967
        ],
    }


async def async_proxy_command(
    *,
    contract_address: spec.Address,
    verbose: bool,
    block: spec.BlockNumberReference,
) -> None:

    contract_address = await evm.async_resolve_address(
        contract_address,
        block=block,
    )

    if verbose:

        styles = cli.get_cli_styles()

        proxy_metadata = await evm.async_get_proxy_metadata(
            contract_address=contract_address,
            block=block,
        )

        if proxy_metadata['implementation'] is None:
            toolstr.print(
                'no proxy detected for '
                + toolstr.add_style(contract_address, styles['metavar']),
            )
            return

        toolstr.print_text_box(
            'Proxy Summary for ' + str(contract_address),
            style=styles['title'],
        )
        rows: typing.Sequence[typing.Sequence[typing.Any]] = [
            ['proxy', toolstr.add_style(contract_address, styles['metavar'])],
            [
                'implementation',
                toolstr.add_style(
                    proxy_metadata['implementation'], styles['metavar']
                ),
            ],
            ['block', block],
            ['proxy_type', proxy_metadata['proxy_type']],
        ]
        toolstr.print_table(
            rows,
            column_justify=['right', 'left'],
            compact=2,
            border=styles['comment'],
            column_styles=[styles['option'], styles['description'] + ' bold'],
            indent=4,
        )

    else:
        proxy_address = await evm.async_get_proxy_implementation(
            contract_address
        )
        if proxy_address is None:
            print('[no proxy address detected]')
        else:
            print(proxy_address)



"""type information that should be available to other modules

(type definitions are only loaded while type checking via typing.TYPE_CHECKING)
"""

from __future__ import annotations


config_keys = (
    'config_spec_version',
    'data_dir',
    'providers',
    'networks',
    'default_network',
    'default_providers',
    'db_configs',
    'log_rpc_calls',
    'log_sql_queries',
    'cli_color_theme',
    'cli_chart_charset',
    'context_cache_rules',
)

config_int_subkeys = (
    'networks',
    'default_providers',
    'network_cache_configs',
)

context_keys = (
    'network',
    'provider',
    'cache',
)

schema_names = (
    #
    # admin schemas
    'schema_versions',
    #
    # generic schemas
    '4byte',
    'coingecko',
    #
    # network schemas
    'block_timestamps',
    'block_gas',
    'blocks',
    'contract_abis',
    'contract_creation_blocks',
    'dex_pools',
    'erc20_metadata',
    'events',
    'transactions',
    'chainlink',
)

network_schema_names = (
    'block_timestamps',
    'block_gas',
    'blocks',
    'contract_abis',
    'contract_creation_blocks',
    'dex_pools',
    'erc20_metadata',
    'events',
    'transactions',
    'chainlink',
)

# context_cache_keys: tuple[Literal['backend', 'read', 'write'], ...] = (
#     'backend',
#     'read',
#     'write',
# )

rpc_block_keys = (
    'base_fee_per_gas',
    'difficulty',
    'extra_data',
    'gas_limit',
    'gas_used',
    'hash',
    'logs_bloom',
    'miner',
    'mix_hash',
    'nonce',
    'number',
    'parent_hash',
    'receipts_root',
    'sha3_uncles',
    'size',
    'state_root',
    'timestamp',
    'total_difficulty',
    'transactions',
    'transactions_root',
    'uncles',
)

db_block_keys = (
    'number',
    'hash',
    'timestamp',
    'miner',
    'extra_data',
    'base_fee_per_gas',
    'gas_limit',
    'gas_used',
)

block_number_names = ['latest', 'earliest', 'pending']

block_keys_standardized = {
    'uncle_hashes': 'sha3_uncles',
    'coinbase': 'miner',
    'author': 'miner',
    'beneficiary': 'miner',
    'transaction_root': 'transactions_root',
    'bloom': 'logs_bloom',
    'block_number': 'number',
    'proof_of_work': 'mix_hash',
}


#
# # transactions
#

transaction_keys_legacy = (
    'nonce',
    'gas_price',
    'gas',
    'to',
    'value',
    'input',
    'v',
    'r',
    's',
)

transaction_keys_eip2930 = (
    'chain_id',
    'nonce',
    'gas_price',
    'gas',
    'to',
    'value',
    'input',
    'access_list',
    'v',
    'r',
    's',
)

transaction_keys_eip2930_camel = (
    'chainId',
    'nonce',
    'gasPrice',
    'gas',
    'to',
    'value',
    'input',
    'accessList',
    'v',
    'r',
    's',
)

transaction_keys_eip1559 = (
    'chain_id',
    'nonce',
    'max_priority_fee_per_gas',
    'max_fee_per_gas',
    'gas',
    'to',
    'value',
    'input',
    'access_list',
    'v',
    'r',
    's',
)

transaction_keys_eip1559_camel = (
    'chainId',
    'nonce',
    'maxPriorityFeePerGas',
    'maxFeePerGas',
    'gas',
    'to',
    'value',
    'input',
    'accessList',
    'v',
    'r',
    's',
)

rpc_transaction_keys = {
    'access_list',
    'block_hash',
    'block_number',
    'chain_id',
    'from',
    'gas',
    'gas_price',
    'hash',
    'input',
    'max_fee_per_gas',
    'max_priority_fee_per_gas',
    'nonce',
    'r',
    's',
    'to',
    'transaction_index',
    'type',
    'v',
    'value',
}

db_transaction_keys = {
    'hash',
    'block_number',
    'transaction_index',
    'to_address',
    'from_address',
    'value',
    'input',
    'nonce',
    'transaction_type',
    'status',
    'gas_used',
    'gas_limit',
    'gas_priority',
    'gas_price',
    'gas_price_max',
    'access_list',
}

# removed_rpc_keys = {'v', 'r', 's', 'chain_id'}


#
# # providers
#

provider_keys = [
    'url',
    'name',
    'network',
    'protocol',
    'session_kwargs',
    'chunk_size',
    'convert_reverts_to_none',
    'disable_batch_requests',
]

default_provider_settings = {
    # these must be particularly specified
    # 'url',
    # 'network',
    # 'protocol',
    #
    'name': '',
    'session_kwargs': {},
    'chunk_size': None,
    'convert_reverts_to_none': False,
    'disable_batch_requests': True,
}


#
# # events
#

event_index_fields = [
    'block_number',
    'transaction_index',
    'log_index',
]

decoded_event_fields = [
    'transaction_hash',
    'contract_address',
    'event_hash',
]

encoded_event_fields = [
    'transaction_hash',
    'contract_address',
    'event_hash',
    'topic1',
    'topic2',
    'topic3',
    'unindexed',
]




from __future__ import annotations

import toolsql

from ctc import config
from ctc import spec
from ... import intake_utils
from ... import management

from . import contract_creation_blocks_statements


async def async_intake_contract_creation_block(
    contract_address: spec.Address,
    block: int,
    *,
    context: spec.Context = None,
) -> None:

    if not management.get_active_schemas().get('contract_creation_blocks'):
        return

    db_config = config.get_context_db_config(
        schema_name='contract_creation_blocks',
        context=context,
    )
    async with toolsql.async_connect(db_config) as conn:
        confirmed = await intake_utils.async_is_block_fully_confirmed(
            block=block, context=context, conn=conn
        )
        if not confirmed:
            return
        await contract_creation_blocks_statements.async_upsert_contract_creation_block(
            conn=conn,
            block_number=block,
            address=contract_address,
            context=context,
        )




from __future__ import annotations

import typing

from ctc import spec
from . import config_read


_config_overrides: spec.PartialConfig = {}


def set_config_override(key: str, value: typing.Any) -> None:
    _config_overrides[key] = value  # type: ignore
    config_read.reset_config_cache()


def get_config_overrides() -> spec.PartialConfig:
    return _config_overrides


def clear_config_override(key: str) -> None:
    del _config_overrides[key]  # type: ignore
    config_read.reset_config_cache()


def clear_all_config_overrides() -> None:
    for key in list(_config_overrides.keys()):
        del _config_overrides[key]  # type: ignore
    config_read.reset_config_cache()




from __future__ import annotations

import toolcli

from ctc import evm
from ctc import rpc


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_bytecode_command,
        'help': 'get raw bytecode stored at address',
        'args': [
            {'name': 'address', 'help': 'address where bytecode is stored'},
        ],
        'examples': [
            '0x6b175474e89094c44da98b954eedeac495271d0f',
        ],
    }


async def async_bytecode_command(address: str) -> None:
    address = await evm.async_resolve_address(address)
    bytecode = await rpc.async_eth_get_code(address)
    print(bytecode)



from __future__ import annotations

import typing

from ctc import spec


erc4626_event_abis: typing.Mapping[str, spec.EventABI] = {
    'Deposit': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'caller',
                'type': 'address',
            },
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'owner',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'assets',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'shares',
                'type': 'uint256',
            },
        ],
        'name': 'Deposit',
        'type': 'event',
    },
    'Withdraw': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'caller',
                'type': 'address',
            },
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'receiver',
                'type': 'address',
            },
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'owner',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'assets',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'shares',
                'type': 'uint256',
            },
        ],
        'name': 'Withdraw',
        'type': 'event',
    },
}


# 18 function abis
# - 4 are write functions (deposit, mint, redeem, withdraw)
# - 1 function is metadata (assets)
# - 2 functions are derived from erc20 (balanceOf, totalSupply)
# - 11 functions otherwise
erc4626_function_abis: typing.Mapping[str, spec.FunctionABI] = {
    'asset': {
        'inputs': [],
        'name': 'asset',
        'outputs': [
            {
                'internalType': 'contract ERC20',
                'name': '',
                'type': 'address',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'balanceOf': {
        'inputs': [
            {
                'internalType': 'address',
                'name': 'owner',
                'type': 'address',
            },
        ],
        'name': 'balanceOf',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'convertToAssets': {
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'shares',
                'type': 'uint256',
            },
        ],
        'name': 'convertToAssets',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'convertToShares': {
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'assets',
                'type': 'uint256',
            },
        ],
        'name': 'convertToShares',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'deposit': {
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'assets',
                'type': 'uint256',
            },
            {
                'internalType': 'address',
                'name': 'receiver',
                'type': 'address',
            },
        ],
        'name': 'deposit',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': 'shares',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'maxDeposit': {
        'inputs': [
            {
                'internalType': 'address',
                'name': 'recipient',
                'type': 'address',
            },
        ],
        'name': 'maxDeposit',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'maxMint': {
        'inputs': [
            {
                'internalType': 'address',
                'name': 'recipient',
                'type': 'address',
            },
        ],
        'name': 'maxMint',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'maxRedeem': {
        'inputs': [
            {
                'internalType': 'address',
                'name': 'owner',
                'type': 'address',
            },
        ],
        'name': 'maxRedeem',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'maxWithdraw': {
        'inputs': [
            {
                'internalType': 'address',
                'name': 'owner',
                'type': 'address',
            },
        ],
        'name': 'maxWithdraw',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'mint': {
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'shares',
                'type': 'uint256',
            },
            {
                'internalType': 'address',
                'name': 'receiver',
                'type': 'address',
            },
        ],
        'name': 'mint',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': 'assets',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'previewDeposit': {
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'assets',
                'type': 'uint256',
            },
        ],
        'name': 'previewDeposit',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'previewMint': {
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'shares',
                'type': 'uint256',
            },
        ],
        'name': 'previewMint',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'previewRedeem': {
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'shares',
                'type': 'uint256',
            },
        ],
        'name': 'previewRedeem',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'previewWithdraw': {
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'assets',
                'type': 'uint256',
            },
        ],
        'name': 'previewWithdraw',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'redeem': {
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'shares',
                'type': 'uint256',
            },
            {
                'internalType': 'address',
                'name': 'receiver',
                'type': 'address',
            },
            {
                'internalType': 'address',
                'name': 'owner',
                'type': 'address',
            },
        ],
        'name': 'redeem',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': 'assets',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'totalAssets': {
        'inputs': [],
        'name': 'totalAssets',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'totalSupply': {
        'inputs': [],
        'name': 'totalSupply',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'withdraw': {
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'assets',
                'type': 'uint256',
            },
            {
                'internalType': 'address',
                'name': 'receiver',
                'type': 'address',
            },
            {
                'internalType': 'address',
                'name': 'owner',
                'type': 'address',
            },
        ],
        'name': 'withdraw',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': 'shares',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
}



from __future__ import annotations

import typing

from ctc import spec


vault = '0xba12222222228d8ba445958a75a0704d566bf2c8'


vault_function_abis: typing.Mapping[str, spec.FunctionABI] = {
    'getPool': {
        'inputs': [
            {
                'internalType': 'bytes32',
                'name': 'poolId',
                'type': 'bytes32',
            },
        ],
        'name': 'getPool',
        'outputs': [
            {
                'internalType': 'address',
                'name': '',
                'type': 'address',
            },
            {
                'internalType': 'enum IVault.PoolSpecialization',
                'name': '',
                'type': 'uint8',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'getPoolTokens': {
        'inputs': [
            {
                'internalType': 'bytes32',
                'name': 'poolId',
                'type': 'bytes32',
            },
        ],
        'name': 'getPoolTokens',
        'outputs': [
            {
                'internalType': 'contract IERC20[]',
                'name': 'tokens',
                'type': 'address[]',
            },
            {
                'internalType': 'uint256[]',
                'name': 'balances',
                'type': 'uint256[]',
            },
            {
                'internalType': 'uint256',
                'name': 'lastChangeBlock',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
}

vault_event_abis: typing.Mapping[str, spec.EventABI] = {
    'PoolRegistered': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'bytes32',
                'name': 'poolId',
                'type': 'bytes32',
            },
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'poolAddress',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'enum IVault.PoolSpecialization',
                'name': 'specialization',
                'type': 'uint8',
            },
        ],
        'name': 'PoolRegistered',
        'type': 'event',
    },
    'TokensRegistered': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'bytes32',
                'name': 'poolId',
                'type': 'bytes32',
            },
            {
                'indexed': False,
                'internalType': 'contract IERC20[]',
                'name': 'tokens',
                'type': 'address[]',
            },
            {
                'indexed': False,
                'internalType': 'address[]',
                'name': 'assetManagers',
                'type': 'address[]',
            },
        ],
        'name': 'TokensRegistered',
        'type': 'event',
    },
    'Swap': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'bytes32',
                'name': 'poolId',
                'type': 'bytes32',
            },
            {
                'indexed': True,
                'internalType': 'contract IERC20',
                'name': 'tokenIn',
                'type': 'address',
            },
            {
                'indexed': True,
                'internalType': 'contract IERC20',
                'name': 'tokenOut',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'amountIn',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'amountOut',
                'type': 'uint256',
            },
        ],
        'name': 'Swap',
        'type': 'event',
    },
}



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from .. import schema_utils

from ctc import spec


def get_active_schemas() -> typing.Mapping[spec.SchemaName, bool]:
    """return specification of which subset of incoming data to store in db"""
    return {
        # 'block_gas_stats': False,
        'block_timestamps': True,
        'block_gas': True,
        'blocks': True,
        'contract_abis': True,
        'contract_creation_blocks': True,
        'dex_pools': True,
        'erc20_metadata': True,
        'events': True,
        'transactions': True,
        # 'erc20_state': False,
        '4byte': True,
        'chainlink': True,
        'schema_versions': True,
    }


def get_active_timestamp_schema() -> spec.NetworkSchemaName | None:
    active_schemas = get_active_schemas()
    if active_schemas['block_timestamps']:
        return 'block_timestamps'
    elif active_schemas['blocks']:
        return 'blocks'
    else:
        return None



from __future__ import annotations

import toolcli

from ctc import config


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': path_command,
        'help': 'print config path',
        'examples': [''],
        'hidden': True,
    }


def path_command() -> None:
    print(config.get_config_path(raise_if_dne=False))



from __future__ import annotations

import typing

import toolcli
import toolstr
import tooltime

from ctc import cli
from ctc import evm
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_address_transactions_command,
        'help': 'output all transactions from address',
        'args': [
            {'name': 'address', 'help': 'get transactions from this address'},
            {
                'name': ['-v', '--verbose'],
                'help': 'display additional information',
                'action': 'store_true',
            },
        ],
        'examples': {
            '0xd8da6bf26964af9d7eed9e03e53415d37aa96045': {'long': True},
            '0xd8da6bf26964af9d7eed9e03e53415d37aa96045 -v': {'long': True},
        },
    }


async def async_address_transactions_command(
    address: spec.Address,
    verbose: bool,
) -> None:

    styles = cli.get_cli_styles()
    toolstr.print(
        'fetching transactions from address',
        toolstr.add_style(address, styles['metavar']),
    )
    print()
    print('this may take some time...')
    print()

    # get row limit
    if verbose:
        limit_rows = None
    else:
        limit_rows = toolcli.get_n_terminal_rows() - 8

    address = await evm.async_resolve_address(address)
    transactions = await evm.async_get_transactions_from_address(address)

    # get block numbers
    if verbose:
        block_numbers = [
            transaction['block_number'] for transaction in transactions
        ]
        block_timestamps = await evm.async_get_block_timestamps(block_numbers)

    rows = []
    for t, transaction in enumerate(transactions):

        row: list[typing.Any] = [transaction['nonce']]
        row.append(str(transaction['block_number']))
        if verbose:
            timestamp = block_timestamps[t]
            row.append(tooltime.timestamp_to_iso_pretty(timestamp))
        row.append(transaction['hash'])

        rows.append(row)

    if verbose:
        labels = ['#', 'block', 'timestamp', 'transaction hash']
    else:
        labels = ['#', 'block', 'transaction hash']

    toolstr.print_table(
        rows,
        labels=labels,
        compact=2,
        limit_rows=limit_rows,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            '#': styles['option'],
            'block': styles['description'],
            'timestamp': styles['description'],
            'transaction hash': styles['metavar'],
        },
    )



from __future__ import annotations

from ctc import spec

from .. import rpc_constructors
from .. import rpc_digestors
from .. import rpc_request


def sync_eth_gas_price(
    *, context: spec.Context = None, decode_response: bool = True
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_gas_price()
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_gas_price(
        response=response,
        decode_response=decode_response,
    )


def sync_eth_accounts(
    *,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_accounts()
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_accounts(response=response)


def sync_eth_sign(
    address: spec.Address,
    message: str,
    *,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_sign(
        address=address,
        message=message,
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_sign(response=response)


def sync_eth_sign_transaction(
    from_address: spec.Address,
    data: str,
    *,
    to_address: spec.Address | None = None,
    gas: int | None = None,
    gas_price: int | None = None,
    value: int | None = None,
    nonce: str | None = None,
    context: spec.Context = None,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_sign_transaction(
        from_address=from_address,
        data=data,
        to_address=to_address,
        gas=gas,
        gas_price=gas_price,
        value=value,
        nonce=nonce,
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_sign_transaction(
        response=response,
        snake_case_response=snake_case_response,
    )


def sync_eth_send_transaction(
    from_address: spec.Address,
    data: str,
    *,
    to_address: spec.Address | None = None,
    gas: int | None = None,
    gas_price: int | None = None,
    value: int | None = None,
    nonce: str | None = None,
    context: spec.Context = None,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_send_transaction(
        from_address=from_address,
        data=data,
        to_address=to_address,
        gas=gas,
        gas_price=gas_price,
        value=value,
        nonce=nonce,
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_send_transaction(
        response=response,
        snake_case_response=snake_case_response,
    )


def sync_eth_send_raw_transaction(
    data: str, *, context: spec.Context = None
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_send_raw_transaction(data=data)
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_send_raw_transaction(response=response)




from __future__ import annotations

import typing

from ctc import spec

if typing.TYPE_CHECKING:

    from typing_extensions import TypedDict

    import toolsql

    class ChainlinkFeed(TypedDict):
        address: str
        name: str
        deviation: str
        heartbeat: str
        decimals: int
        asset: str
        asset_type: str
        status: str

    class _FeedAggregatorUpdate(TypedDict):
        feed: spec.Address
        aggregator: spec.Address
        block_number: int


chainlink_schema: toolsql.DBSchemaShorthand = {
    'tables': {
        'chainlink_feeds': {
            'columns': [
                {'name': 'address', 'type': 'Text', 'primary': True},
                {'name': 'name', 'type': 'Text', 'index': True},
                {'name': 'deviation', 'type': 'Text'},
                {'name': 'heartbeat', 'type': 'Text'},
                {'name': 'decimals', 'type': 'Integer'},
                {'name': 'asset', 'type': 'Text', 'index': True},
                {'name': 'asset_type', 'type': 'Text', 'index': True},
                {'name': 'status', 'type': 'Text'},
            ],
        },
        'chainlink_aggregator_updates': {
            'columns': [
                {'name': 'feed', 'type': 'Text', 'primary': True},
                {'name': 'aggregator', 'type': 'Text', 'primary': True},
                {'name': 'block_number', 'type': 'Integer', 'primary': True},
            ],
        },
    },
}



from __future__ import annotations

import os
import re
import tempfile

import pytest

import ctc


def get_readme_examples():

    path = get_readme_path()
    if not os.path.exists(path):
        return {'python': [], 'bash': []}

    else:

        with open(get_readme_path(), 'r') as f:
            content = f.read()

        return {
            'python': re.findall('```python\n(.+?)```\n', content, re.DOTALL),
            'bash': re.findall('```bash\n(.+?)```\n', content, re.DOTALL),
        }


def get_readme_path():
    root_dir = os.path.dirname(os.path.dirname(ctc.__path__[0]))
    return os.path.join(root_dir, 'README.md')


readme_examples = get_readme_examples()


@pytest.mark.parametrize('example', readme_examples['python'])
async def test_readme_python_examples(example):
    example_lines = example.split('\n')
    code = 'async def f_example():\n' + '    ' + '\n    '.join(example_lines)
    exec(code)
    await locals()['f_example']()


@pytest.mark.parametrize('example', readme_examples['bash'])
async def test_readme_bash_examples(example):
    tempdir = tempfile.mkdtemp()
    os.chdir(tempdir)

    script = '#!/bin/bash\n\nset -e\n\n' + example
    with open('script.sh', 'w') as f:
        f.write(script)

    import subprocess
    subprocess.check_call('bash script.sh', shell=True)



import toolsql

from ctc import db

import conftest


example_data = [
    {
        'block_number': 14000000,
        'timestamp': 1642114795,
    },
    {
        'block_number': 14000001,
        'timestamp': 1642114800,
    },
    {
        'block_number': 14000002,
        'timestamp': 1642114824,
    },
    {
        'block_number': 14000003,
        'timestamp': 1642114825,
    },
    {
        'block_number': 14000004,
        'timestamp': 1642114850,
    },
    {
        'block_number': 14000005,
        'timestamp': 1642114852,
    },
    {
        'block_number': 14000006,
        'timestamp': 1642114865,
    },
    {
        'block_number': 14000007,
        'timestamp': 1642114881,
    },
    {
        'block_number': 14000008,
        'timestamp': 1642114895,
    },
    {
        'block_number': 14000009,
        'timestamp': 1642114917,
    },
]


async def test_block_timestamps_db():
    db_config = conftest.get_test_db_config()
    db_schema = db.get_prepared_schema(
        schema_name='block_timestamps',
        context=dict(network='ethereum'),
    )
    toolsql.create_db(
        db_config=db_config,
        db_schema=db_schema,
        if_not_exists=True,
        confirm=True,
    )

    # insert data one-by-one
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            await db.async_upsert_block_timestamp(conn=conn, **datum)

    # get data individually
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            timestamp = await db.async_select_block_timestamp(
                conn=conn,
                block_number=datum['block_number'],
            )
            assert timestamp == datum['timestamp']

    # get data collectively
    all_blocks = [datum['block_number'] for datum in example_data]
    all_timestamps = [datum['timestamp'] for datum in example_data]
    async with toolsql.async_connect(db_config) as conn:
        stored_timestamps = await db.async_select_block_timestamps(
            conn=conn,
            block_numbers=all_blocks,
        )
        assert set(stored_timestamps) == set(all_timestamps)

    # delete entries one by one
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            await db.async_delete_block_timestamp(
                conn=conn,
                block_number=datum['block_number'],
            )

    # ensure all entries deleted
    async with toolsql.async_connect(db_config) as conn:
        stored_timestamps = await db.async_select_block_timestamps(
            conn=conn,
            block_numbers=all_blocks,
        )
        assert set(stored_timestamps) == {None}

    # insert data again
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            await db.async_upsert_block_timestamp(conn=conn, **datum)

    # delete entries all at once
    async with toolsql.async_connect(db_config) as conn:
        await db.async_delete_block_timestamps(
            conn=conn,
            block_numbers=all_blocks,
        )

    # ensure all entries deleted
    async with toolsql.async_connect(db_config) as conn:
        stored_timestamps = await db.async_select_block_timestamps(
            conn=conn,
            block_numbers=all_blocks,
        )
        assert set(stored_timestamps) == {None}



from __future__ import annotations

import typing

from . import fourbyte_db
from . import fourbyte_spec


#
# # database population functions
#


async def async_build_function_signatures_dataset(
    signature_data: typing.Sequence[fourbyte_spec.PartialEntry]
    | typing.Sequence[fourbyte_spec.Entry]
    | None = None,
) -> None:

    if signature_data is None:
        # TODO: smarter detection of what is currently in database
        print('building function signatures db from scratch, may take awhile')
        signature_data = await async_scrape_function_signatures()

    await fourbyte_db.async_intake_function_signatures(
        function_signatures=signature_data  # type: ignore
    )


async def async_build_event_signatures_dataset(
    signature_data: typing.Sequence[fourbyte_spec.PartialEntry]
    | typing.Sequence[fourbyte_spec.Entry]
    | None = None,
) -> None:

    if signature_data is None:
        # TODO: smarter detection of what is currently in database
        print('building event signatures db from scratch, may take awhile')
        signature_data = await async_scrape_event_signatures()

    await fourbyte_db.async_intake_event_signatures(
        event_signatures=signature_data,  # type: ignore
    )


#
# # entry scraping functions
#


async def async_scrape_function_signatures(
    *,
    wait_time: typing.Optional[int] = None,
    print_every: typing.Optional[int] = 10000,
    min_id: typing.Optional[int] = None,
) -> typing.Sequence[fourbyte_spec.Entry]:

    return await async_scrape_set(
        url=fourbyte_spec.endpoints['functions'],
        wait_time=wait_time,
        print_every=print_every,
        min_id=min_id,
    )


async def async_scrape_event_signatures(
    *,
    wait_time: typing.Optional[int] = None,
    print_every: typing.Optional[int] = 10000,
    min_id: typing.Optional[int] = None,
) -> typing.Sequence[fourbyte_spec.Entry]:

    return await async_scrape_set(
        url=fourbyte_spec.endpoints['events'],
        wait_time=wait_time,
        print_every=print_every,
        min_id=min_id,
    )


async def async_scrape_set(
    url: str,
    *,
    wait_time: typing.Optional[int] = None,
    print_every: typing.Optional[int] = 10000,
    min_id: typing.Optional[int] = None,
) -> typing.Sequence[fourbyte_spec.Entry]:

    import aiohttp

    results = []

    next_print = None
    if print_every is not None:
        next_print = print_every

    async with aiohttp.ClientSession() as session:
        while True:

            # get page
            async with session.get(url) as response_object:
                response = await response_object.json()
                results.extend(response['results'])

            # scrape only until min_id is reached
            if min_id is not None:
                min_result_id = min(
                    result['id'] for result in response['results']
                )
                if min_result_id < min_id:
                    break

            # get next url
            url = response['next']
            if url is None:
                break

            # print verbose message
            if print_every is not None and next_print is not None:
                if len(results) > next_print:
                    while next_print < len(results):
                        print('scraped', next_print, 'results')
                        next_print += print_every

            # wait between responses
            if wait_time is not None:
                import asyncio

                await asyncio.sleep(wait_time)

    return results



"""

TODO:
- tail all log files or select which files to tail
    - https://unix.stackexchange.com/questions/149017/how-to-tail-multiple-files-using-tail-0f-in-linux-aix
"""

from __future__ import annotations

import os
import subprocess

import toolcli

from ctc import config


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': log_command,
        'help': 'display logs',
        'args': [],
        'examples': {'': {'description': 'tail logs', 'runnable': False}},
    }


def log_command() -> None:
    rpc_requests_log_path = config.get_rpc_requests_log_path()
    print('watching log file:', rpc_requests_log_path)
    try:
        n_lines = os.get_terminal_size().lines - 3
        subprocess.call(
            ['tail', '-n', str(n_lines), '-f', rpc_requests_log_path]
        )
    except KeyboardInterrupt:
        print()



from __future__ import annotations

import typing

from ctc import evm
from ctc import spec
from . import multicall_spec


def get_call_contract(call: multicall_spec.Call) -> spec.Address:
    if isinstance(call, dict):
        return call['contract']
    elif isinstance(call, (list, tuple)):
        result = call[0]
        if not isinstance(result, str):
            raise Exception('bad format for call')
        return result
    else:
        raise Exception('unknown call format')


async def async_encode_call(
    call: multicall_spec.Call,
    *,
    context: spec.Context = None,
) -> tuple[spec.Address, spec.BinaryData]:
    contract = get_call_contract(call)
    call_data = await async_encode_call_data(call=call, context=context)
    return (contract, call_data)


async def async_encode_call_data(
    call: multicall_spec.Call,
    *,
    context: spec.Context = None,
) -> spec.BinaryData:

    # parse components
    if isinstance(call, dict):
        if 'call_data' in call:
            return typing.cast(multicall_spec.EncodedCallDict, call)[
                'call_data'
            ]
        elif 'function' in call:
            call = typing.cast(multicall_spec.UnencodedCallDict, call)
            contract = call['contract']
            function = call['function']
            function_parameters = call.get('function_parameters')

    elif isinstance(call, (list, tuple)):
        if len(call) == 2:
            contract, function = call
            function_parameters = None
        elif len(call) == 3:
            contract, function, function_parameters = call
        else:
            raise Exception('unknown call format')

    else:
        raise Exception('unknown call format')

    # get abi
    if isinstance(function, dict):
        function_abi = function
    elif isinstance(function, str):
        function_abi = await evm.async_get_function_abi(
            contract_address=contract,
            function_name=function,
            context=context,
        )
    else:
        raise Exception('could not determine function_abi')

    # encode
    encoded_data = evm.encode_call_data(
        function_abi=function_abi,
        parameters=function_parameters,
    )
    return evm.to_binary(encoded_data)


async def async_decode_call_output(
    call: multicall_spec.Call,
    encoded_output: spec.BinaryData,
    *,
    context: spec.Context = None,
) -> typing.Any:
    function_abi = await async_get_call_function_abi(call, context=context)
    return evm.decode_function_output(
        encoded_output=encoded_output,
        function_abi=function_abi,
    )


async def async_get_call_function_abi(
    call: multicall_spec.Call,
    *,
    context: spec.Context = None,
) -> spec.FunctionABI:

    function: spec.FunctionABI | str | None = None
    call_data: spec.BinaryData | None = None
    if isinstance(call, dict):
        if 'function' in call:
            function = typing.cast(multicall_spec.UnencodedCallDict, call)[
                'function'
            ]
        elif 'call_data' in call:
            call_data = typing.cast(multicall_spec.EncodedCallDict, call)[
                'call_data'
            ]
        else:
            raise Exception('unknown call format')
    elif isinstance(call, (list, tuple)):
        if isinstance(call[1], dict):
            if typing.TYPE_CHECKING:
                return typing.cast(spec.FunctionABI, call[1])
            else:
                return call[1]
        elif isinstance(call[1], str) and not call[1].startswith('0x'):
            function = call[1]
        else:
            call_data = call[1]
    else:
        raise Exception('unknown call format')

    if function is not None:
        if isinstance(function, dict):
            return function
        elif isinstance(function, str):
            return await evm.async_get_function_abi(
                contract_address=get_call_contract(call),
                function_name=function,
                context=context,
            )
        else:
            raise Exception('unknown call format')

    elif call_data is not None:
        call_data = evm.to_hex(call_data)
        function_selector = call_data[:10]
        return await evm.async_get_function_abi(
            contract_address=get_call_contract(call),
            function_selector=function_selector,
            context=context,
        )

    else:
        raise Exception('unknown call format')



from .chainlink_data import *

from .chainlink_aggregators import *
from .chainlink_helpers import *
from .chainlink_feed_metadata import *
from .chainlink_registry import *
from .chainlink_spec import *
from .chainlink_summary import *



from .context_utils import *
from .config_overrides import *
from .config_read import *
from .config_spec import *
from .config_values import *
from .config_write import *

# do not import
# - config_defaults
# - setup_utils




from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import toolsql

from typing_extensions import Literal
from typing_extensions import TypedDict

from . import rpc_types


BackendType = Literal['filesystem', 'rpc', 'db', 'rest', 'hybrid']


class DataSource(TypedDict, total=False):
    """rpc_typesifies a location to use for retrieving data

    - if source parameters are missing, system defaults are used
    - for hybrid sources, will try sources in order until one succeeds
        - will insert result into earlier source if hybrid_backfill=True
    """

    # backend type
    backend: BackendType

    # source parameters
    db_config: toolsql.DBConfig
    rest_endpoint: dict[str, typing.Any]
    filesystem_root: str
    provider: rpc_types.ProviderReference | None

    # hybrid parameters
    hybrid_order: typing.Sequence['LeafDataSource']
    hybrid_backfill: bool


class LeafDataSource(TypedDict, total=False):

    # backend type
    backend: BackendType

    # source parameters
    db_config: toolsql.DBConfig
    rest_endpoint: str
    filesystem_root: str
    provider: rpc_types.ProviderReference



from __future__ import annotations

import typing

from ctc import evm
from ctc import spec
from .. import rpc_format


#
# # helper decoding functions
#


def _decode_single_trace(trace: typing.Any) -> None:
    trace['action'] = rpc_format.decode_response(
        trace['action'],
        ['gas', 'value'],
    )
    if trace['result'] is not None:
        trace['result'] = rpc_format.decode_response(
            trace['result'],
            ['gasUsed'],
        )


def _decode_trace_list(traces: typing.Any) -> None:
    for trace in traces:
        _decode_single_trace(trace)


def _decode_trace_collection(traces: typing.Any) -> None:
    if traces['trace'] is not None:
        _decode_trace_list(traces['trace'])
    if traces.get('stateDiff') is not None:
        _decode_state_diff(traces['stateDiff'])


def _decode_state_diff(state_diff: spec.StateDiffTrace) -> None:

    for diff in state_diff.values():
        for key in ['balance', 'nonce']:
            subdiff = diff[key]  # type: ignore
            if isinstance(subdiff, dict):
                if '+' in subdiff:
                    subdiff['+'] = evm.binary_convert(subdiff['+'], 'integer')
                elif '*' in subdiff:
                    subdiff['*']['to'] = evm.binary_convert(
                        subdiff['*']['to'], 'integer'
                    )
                    subdiff['*']['from'] = evm.binary_convert(
                        subdiff['*']['from'], 'integer'
                    )
                elif '-' in subdiff:
                    subdiff['-'] = evm.binary_convert(subdiff['-'], 'integer')
                else:
                    raise Exception('unknown subdiff format')


def _snake_case_single_trace(trace: typing.Any) -> typing.Any:
    trace['action'] = rpc_format.keys_to_snake_case(trace['action'])
    if trace['result'] is not None:
        trace['result'] = rpc_format.keys_to_snake_case(trace['result'])
    return rpc_format.keys_to_snake_case(trace)


def _snake_case_trace_list(traces: typing.Any) -> typing.Any:
    traces = [rpc_format.keys_to_snake_case(trace) for trace in traces]
    return [_snake_case_single_trace(trace) for trace in traces]


def _snake_case_trace_collection(traces: typing.Any) -> None:

    if 'transactionHash' in traces:
        traces['transaction_hash'] = traces.pop('transactionHash')
    if 'vmTrace' in traces:
        traces['vm_trace'] = traces.pop('vmTrace')
    if 'stateDiff' in traces:
        traces['state_diff'] = traces.pop('stateDiff')

    if traces['trace'] is not None:
        traces['trace'] = _snake_case_trace_list(traces['trace'])


def _snake_case_debug_block_trace(trace: typing.Any) -> spec.DebugBlockTrace:
    return [_snake_case_debug_transaction_trace(tx_trace) for tx_trace in trace]


def _snake_case_debug_transaction_trace(
    trace: typing.Any,
) -> spec.DebugTransactionTrace:
    return {
        'struct_logs': _snake_case_debug_struct_log_traces(trace['structLogs']),
        'gas': trace['gas'],
        'failed': trace['failed'],
        'return_value': trace['returnValue'],
    }


def _snake_case_debug_struct_log_traces(
    traces: typing.Any,
) -> typing.Sequence[spec.DebugTraceStructLog]:
    for trace in traces:
        trace['gas_cost'] = trace.pop('gasCost')
    return traces  # type: ignore


#
# # transactions
#


def digest_trace_transaction(
    response: spec.RpcSingularResponse,
    *,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:

    if decode_response:
        _decode_trace_list(response)

    if snake_case_response:
        response = _snake_case_trace_list(response)

    return response


def digest_trace_replay_transaction(
    response: spec.RpcSingularResponse,
    *,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:

    if decode_response:
        _decode_trace_collection(response)

    if snake_case_response:
        _snake_case_trace_collection(response)

    return response


def digest_trace_raw_transaction(
    response: spec.RpcSingularResponse,
    *,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    return response


#
# # calls
#


def digest_trace_call(
    response: spec.RpcSingularResponse,
    *,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:

    if decode_response:
        _decode_trace_collection(response)

    if snake_case_response:
        _snake_case_trace_collection(response)

    return response


def digest_trace_call_many(
    response: spec.RpcSingularResponse,
    *,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:

    if decode_response:
        for subresponse in response:
            _decode_trace_collection(subresponse)

    if snake_case_response:
        for subresponse in response:
            _snake_case_trace_collection(subresponse)

    return response


#
# # trace ranges
#


def digest_trace_get(
    response: spec.RpcSingularResponse,
    *,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:

    if decode_response:
        _decode_single_trace(response)

    if snake_case_response:
        response = _snake_case_single_trace(response)

    return response


def digest_trace_filter(
    response: spec.RpcSingularResponse,
    *,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:

    if decode_response:
        _decode_trace_list(response)

    if snake_case_response:
        response = _snake_case_trace_list(response)

    return response


def digest_trace_block(
    response: spec.RpcSingularResponse,
    *,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:

    if decode_response:
        _decode_trace_list(response)

    if snake_case_response:
        response = _snake_case_trace_list(response)

    return response


def digest_trace_replay_block_transactions(
    response: spec.RpcSingularResponse,
    *,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:

    if decode_response:
        for subtrace in response:
            _decode_trace_collection(subtrace)

    if snake_case_response:
        for subtrace in response:
            _snake_case_trace_collection(subtrace)

    return response


#
# # debug
#


def digest_debug_trace_call(
    response: spec.RpcSingularResponse,
    *,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:

    if snake_case_response:
        response = _snake_case_debug_transaction_trace(response)

    return response


def digest_debug_trace_call_many(
    response: spec.RpcSingularResponse,
    *,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:

    if snake_case_response:
        response = [
            _snake_case_debug_transaction_trace(subresponse)
            for subresponse in response
        ]

    return response


def digest_debug_trace_transaction(
    response: spec.RpcSingularResponse,
    *,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:

    if snake_case_response:
        response = _snake_case_debug_transaction_trace(response)

    return response


def digest_debug_trace_block_by_number(
    response: spec.RpcSingularResponse,
    *,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:

    if snake_case_response:
        response = _snake_case_debug_block_trace(response)

    return response


def digest_debug_trace_block_by_hash(
    response: spec.RpcSingularResponse,
    *,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:

    if snake_case_response:
        response = _snake_case_debug_block_trace(response)

    return response




import pytest

from ctc.defi import dex_utils


#
# # test get dex pools
#

example_pool_queries = [
    {
        'query': {
            'assets': ['0x5f98805a4e8be255a32880fdec7f6728c6568ba0'],
        },
        'results': [
            '0xb0f75e97a114a4eb4a425edc48990e6760726709',
            '0xede4efcc5492cf41ed3f0109d60bc0543cfad23a',
            '0x5a172919fff040196b36b238b9e2e7e6d33f616a',
            '0xc4451498f950b8b3abd9a815cf221a8e64791388',
            '0x1b46e4b0791c9383b73b64aabc371360a031a83f',
            '0xb39362c3d5ac235fe588b0b83ed7ac87241039cb',
            '0xd652c40fbb3f06d6b58cb9aa9cff063ee63d465d',
            '0xed279fdd11ca84beef15af5d39bb4d4bee23f0ca',
            '0x497ce58f34605b9944e6b15ecafe6b001206fd25',
            '0xa0d35faead5299bf18efbb5defd1ec6d4ab4ef3b',
            '0x780d605cd56043664478292705bf8027f215942f',
            '0x46e4d8a1322b9448905225e52f914094dbd6dddf',
            '0x279ca79d5fb2490721512c8ae4767e249d75f41b',
            '0xfdf12d1f85b5082877a6e070524f50f6c84faa6b',
            '0x20062da66f4d36942313c3395e5936cf04e8320a',
            '0xf20ef17b889b437c151eb5ba15a47bfc62bff469',
            '0x477a61f69d468fb61224a33b7b0d4de647b4f704',
            '0x6255550a53fd12b7131c0829dc439c5e02eebca8',
            '0x16980c16811bde2b3358c1ce4341541a4c772ec9',
            '0xbcc2adccd4de0b2656179f729c98c39dcd68c84d',
            '0xa31a285832edbffef40d271a4ff1180bdd30237f',
            '0xf1ddc084500fc774bb835f412aea908259b55ce4',
            '0xd251dff33e31bb98d5587e5b1004ff01a5a41289',
            '0xaaadbb0d64f6d25731306f11e9ee94b486d53e43',
            '0x159109cacaf431005099b73ef8b66069682a66b4',
            '0x4e0924d3a751be199c426d52fb1f2337fa96f736',
            '0x9902affdd3b8ef60304958c60377110c6d6ab1df',
            '0x67e887913b13e280538c169f13d169a659a203de',
            '0x5c09448abfdcb638ea992ce3e45bbc1d7ec7c4cf',
            '0x9663f2ca0454accad3e094448ea6f77443880454',
            '0x2746f8ee2390b33e87c39d9f786b6a004f212064',
            '0x54a0f2d8bd71b793d008266005dfae136500b204',
            '0xd4cd70c4569f1e3661d50004449fd501a386237c',
            '0x2020b31a89fb359ffcb2933f4bba65da6f27d0fe',
            '0xb065c77afc6e1a03b6166ac0fb2f4e84ff6a24d4',
        ],
    },
]


@pytest.mark.parametrize('test', example_pool_queries)
async def test_get_pools(test):
    pools = await dex_utils.async_get_pools(**test['query'])
    assert set(pool['address'] for pool in pools) == set(test['results'])


#
# # test getting DEX balances
#

example_balances = [
    # balancer pool
    {
        'query': {
            'block': 14000000,
            'pool': 'address',
        },
        'result': {
            'token0': 0,
            'token1': 0,
        },
    },
    # curve pool
    {
        'query': {
            'block': 14000000,
            'pool': 'address',
        },
        'result': {
            'token0': 0,
            'token1': 0,
        },
    },
    # uniswap v2 pool
    {
        'query': {
            'block': 14000000,
            'pool': 'address',
        },
        'result': {
            'token0': 0,
            'token1': 0,
        },
    },
    # uniswap v3 pool
    {
        'query': {
            'block': 14000000,
            'pool': 'address',
        },
        'result': {
            'token0': 0,
            'token1': 0,
        },
    },
]


@pytest.mark.parametrize('test', example_balances)
async def test_get_pool_balances(test):
    pool_balances = await dex_utils.async_get_pool_balances(**test['query'])
    assert set(pool_balances.keys()) == set(test['result'])
    for key in pool_balances.keys():
        pool_balances[key] == test['result'][key]


#
# # test getting dex swaps
#

example_swap_queries = [
    {
        'query': {
            'pool': 'address',
            'start_block': 0,
            'end_block': 0,
        },
        'results': {
            'n_swaps': 0,
        },
    },
]


@pytest.mark.parametrize('test', example_swap_queries)
async def test_get_pool_swaps(test):
    swaps = await dex_utils.async_get_pool_swaps(**test['query'])
    assert swaps['results']['n_swaps'] == len(swaps)



import pytest

from ctc import rpc


@pytest.mark.asyncio
async def test_eth_gas_price():
    result = await rpc.async_eth_gas_price()
    assert isinstance(result, int)


# deprecated
# @pytest.mark.asyncio
# async def test_eth_accounts():
#     result = await rpc.async_eth_accounts()
#     assert isinstance(result, list)



from __future__ import annotations

import toolsql


erc20_metadata_schema: toolsql.DBSchemaShorthand = {
    'tables': {
        'erc20_metadata': {
            'columns': [
                {
                    'name': 'address',
                    'type': 'Text',
                    'primary': True,
                },
                {
                    'name': 'symbol',
                    'type': 'Text',
                    'index': True,
                },
                {
                    'name': 'decimals',
                    'type': 'Integer',
                },
                {
                    'name': 'name',
                    'type': 'Text',
                    'index': True,
                },
            ],
        },
    },
}



from __future__ import annotations

import typing

from ctc import spec

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict
    from typing_extensions import NotRequired


#
# # calls
#

if typing.TYPE_CHECKING:
    FunctionParameterList = typing.Sequence[typing.Any]
    FunctionParameterMap = typing.Mapping[str, typing.Any]
    FunctionParameters = typing.Union[
        FunctionParameterList, FunctionParameterMap
    ]

    UnencodedCallTuple = typing.Union[
        typing.Tuple[
            spec.Address,
            typing.Union[spec.FunctionABI, str],
        ],
        typing.Tuple[
            spec.Address,
            typing.Union[spec.FunctionABI, str],
            typing.Union[FunctionParameterList, FunctionParameterMap],
        ],
    ]
    UnencodedCallList = typing.Sequence[typing.Any]

    class UnencodedCallDict(TypedDict):
        contract: spec.Address
        function: typing.Union[spec.FunctionABI, str]
        function_parameters: NotRequired[FunctionParameters]

    UnencodedCall = typing.Union[
        UnencodedCallTuple,
        UnencodedCallList,
        UnencodedCallDict,
    ]

    EncodedCallTuple = typing.Tuple[spec.Address, spec.BinaryData]
    EncodedCallList = typing.Sequence[typing.Any]

    class EncodedCallDict(TypedDict):
        contract: spec.Address
        call_data: spec.BinaryData

    EncodedCall = typing.Union[
        EncodedCallTuple, EncodedCallList, EncodedCallDict
    ]

    Call = typing.Union[UnencodedCall, EncodedCall]



from __future__ import annotations

import typing
import warnings

if typing.TYPE_CHECKING:
    import aiohttp

from ctc import config
from ctc import spec
from .. import rpc_provider


_http_sessions: dict[spec.ProviderId, aiohttp.ClientSession] = {}


def sync_send_http(
    request: spec.RpcRequest,
    provider: spec.Provider,
    *,
    n_attempts: int = 8,
) -> str:
    import requests
    import json

    headers = {'Content-Type': 'application/json', 'User-Agent': 'ctc'}
    data = json.dumps(request)
    for attempt in range(n_attempts):
        try:
            response = requests.post(
                provider['url'],
                data=data,
                headers=headers,
            )
            return response.text
        except Exception:
            import time

            time.sleep(0.250)
    else:
        if response is None:
            status = 'None'
        else:
            status = str(response.status_code)
        message = (
            'http rpc request failed after '
            + str(n_attempts)
            + ' retries, status_code = '
            + str(status)
        )
        raise Exception(message)


async def async_send_http(
    request: spec.RpcRequest,
    provider: spec.Provider,
    *,
    n_attempts: int = 8,
) -> str:
    session = get_async_http_session(provider=provider)

    headers = {'User-Agent': 'ctc'}
    response = None
    for attempt in range(n_attempts):
        try:
            async with session.post(
                provider['url'], json=request, headers=headers
            ) as response:
                if response.status != 200:
                    import random

                    t_sleep = 2**attempt + random.random()
                    warnings.warn(
                        'request failed with code '
                        + str(response.status)
                        + ' retrying in '
                        + str(t_sleep)
                        + 's'
                    )
                    import asyncio

                    await asyncio.sleep(t_sleep)
                    continue
                as_text = await response.text()
                return as_text
        except Exception:
            # connection failure
            import asyncio

            await asyncio.sleep(0.250)

    else:
        if response is None:
            status = 'None'
        else:
            status = str(response.status)
        message = (
            'http rpc request failed after '
            + str(n_attempts)
            + ' retries, status_code = '
            + str(status)
        )
        raise Exception(message)


def get_async_http_session(
    provider: spec.Provider, create: bool = True
) -> aiohttp.ClientSession:
    provider_id = rpc_provider._get_provider_id(provider)
    if provider_id not in _http_sessions:
        if create:
            import aiohttp

            kwargs = provider['session_kwargs']
            if kwargs is None:
                kwargs = {}
            kwargs = dict(kwargs)
            kwargs.setdefault('timeout', aiohttp.ClientTimeout(300))
            _http_sessions[provider_id] = aiohttp.ClientSession(
                trust_env=True, **kwargs
            )
        else:
            raise Exception('no session, must create')
    return _http_sessions[provider_id]


async def async_close_http_session(
    context: spec.Context = None,
) -> None:
    import asyncio

    if len(_http_sessions) == 0:
        return

    if context is None:
        for key, session in list(_http_sessions.items()):
            await asyncio.sleep(0)
            await session.close()
            del _http_sessions[key]

    else:
        provider = config.get_context_provider(context)
        if provider is None:
            raise Exception('no provider available')
        session = get_async_http_session(provider=provider)
        session_keys = [
            key
            for key, value in _http_sessions.items()
            if id(session) == id(value)
        ]
        if len(session_keys) != 1:
            raise Exception('unknown session')
        else:
            session_key = session_keys[0]
        await asyncio.sleep(0)
        await session.close()
        del _http_sessions[session_key]




from __future__ import annotations

import typing

from ctc import spec

from .. import dex_class
from .. import dex_class_utils


async def async_get_pool_balance(
    pool: spec.Address,
    asset: spec.Address,
    *,
    dex: typing.Type[dex_class.DEX] | str | None = None,
    factory: spec.Address | None = None,
    block: spec.BlockNumberReference | None = None,
    normalize: bool = True,
    context: spec.Context = None,
) -> int | float:
    """get balance of particular asset in pool"""

    dex = dex_class_utils.get_dex_class(
        dex=dex, factory=factory, context=context,
    )
    return await dex.async_get_pool_balance(
        pool=pool,
        asset=asset,
        block=block,
        normalize=normalize,
        context=context,
    )


async def async_get_pool_balances(
    pool: spec.Address,
    *,
    dex: typing.Type[dex_class.DEX] | str | None = None,
    factory: spec.Address | None = None,
    normalize: bool = True,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> typing.Mapping[spec.Address, int | float]:
    """get balances of all assets in pool"""

    dex = dex_class_utils.get_dex_class(
        dex=dex, factory=factory, context=context
    )
    return await dex.async_get_pool_balances(
        pool=pool,
        normalize=normalize,
        block=block,
        context=context,
    )


async def async_get_pool_balance_by_block(
    pool: spec.Address,
    asset: spec.Address,
    *,
    blocks: typing.Sequence[spec.BlockNumberReference],
    dex: typing.Type[dex_class.DEX] | str | None = None,
    factory: spec.Address | None = None,
    normalize: bool = True,
    context: spec.Context = None,
) -> typing.Sequence[int | float]:
    """get balances of particular asset in pool at specific blocks"""

    dex = dex_class_utils.get_dex_class(
        dex=dex, factory=factory, context=context,
    )
    return await dex.async_get_pool_balance_by_block(
        pool=pool,
        asset=asset,
        blocks=blocks,
        normalize=normalize,
        context=context,
    )


async def async_get_pool_balances_by_block(
    pool: spec.Address,
    *,
    blocks: typing.Sequence[spec.BlockNumberReference],
    dex: typing.Type[dex_class.DEX] | str | None = None,
    factory: spec.Address | None = None,
    normalize: bool = True,
    context: spec.Context = None,
) -> typing.Mapping[str, typing.Sequence[int | float]]:
    """get balances of all assets in pool at specific blocks"""

    dex = dex_class_utils.get_dex_class(
        dex=dex, factory=factory, context=context
    )
    return await dex.async_get_pool_balances_by_block(
        pool=pool,
        blocks=blocks,
        normalize=normalize,
        context=context,
    )



from __future__ import annotations

import re
import typing


def typehints_formatter(
    annotation: typing.Any,
    sphinx_config: typing.Mapping[str, typing.Any],
) -> str:
    """used for documentation of types by sphinx_autodoc_typehints"""
    output = str(annotation)
    matches = re.findall('[a-zA-Z_\.]+', output)
    for match in matches:
        if '.' in match:
            tokens = match.split('.')
            output = output.replace(match, tokens[-1])
    return output



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import tooltime

from ctc import spec
from ... import block_crud
from . import block_time_search


async def async_get_block_of_timestamp(
    timestamp: tooltime.Timestamp,
    *,
    nary: typing.Optional[int] = None,
    cache: typing.Optional[block_time_search.BlockTimestampSearchCache] = None,
    block_timestamps: typing.Optional[typing.Mapping[int, int]] = None,
    block_timestamp_array: typing.Optional[spec.NumpyArray] = None,
    block_number_array: typing.Optional[spec.NumpyArray] = None,
    verbose: bool = False,
    context: spec.Context = None,
    mode: typing.Literal['<=', '>=', '=='] = '>=',
) -> int:
    """search for the block that corresponds to a given timestamp"""

    if not isinstance(timestamp, int):
        import tooltime

        timestamp = tooltime.timestamp_to_seconds(timestamp)

    if block_timestamps is not None or (
        block_timestamp_array is not None and block_number_array is not None
    ):
        return _get_block_of_timestamp_from_arrays(
            timestamp=timestamp,
            block_timestamp_array=block_timestamp_array,
            block_number_array=block_number_array,
            block_timestamps=block_timestamps,
            verbose=verbose,
        )
    else:

        from ctc import config

        # db
        read_cache, write_cache = config.get_context_cache_read_write(
            schema_name='block_timestamps', context=context
        )
        if read_cache:
            from ctc import db

            block = await db.async_query_timestamp_block(
                timestamp=timestamp,
                mode=mode,
                context=context,
            )
            if block is not None:
                return block

        # rpc node
        return await block_time_search._async_get_block_of_timestamp_from_node(
            timestamp=timestamp,
            nary=nary,
            cache=cache,
            verbose=verbose,
            context=context,
            mode=mode,
        )


def _get_block_of_timestamp_from_arrays(
    timestamp: tooltime.Timestamp,
    *,
    block_timestamp_array: spec.NumpyArray | None = None,
    block_number_array: spec.NumpyArray | None = None,
    block_timestamps: typing.Mapping[int, int] | None = None,
    verbose: bool = False,
) -> int:
    import numpy as np

    if block_timestamp_array is None:
        if block_timestamps is None:
            raise Exception('must specify more arguments')
        block_timestamp_array = np.array(list(block_timestamps.values()))
    if block_number_array is None:
        if block_timestamps is None:
            raise Exception('must specify more arguments')
        block_number_array = np.array(list(block_timestamps.keys()))

    if not isinstance(timestamp, int):
        import tooltime

        timestamp = tooltime.timestamp_to_seconds(timestamp)

    index = np.searchsorted(block_timestamp_array, timestamp)
    result = block_number_array[index]
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_get_block_number_and_time(
    *,
    block_number: typing.Optional[spec.BlockNumberReference] = None,
    block_timestamp: typing.Optional[tooltime.Timestamp] = None,
    context: spec.Context = None,
) -> tuple[int, int]:
    """get block number and timestamp corresponding to given input"""

    if block_timestamp is not None and block_number is not None:
        raise Exception('must specify start_time or block_number')

    if block_number is not None:
        block = await block_crud.async_get_block(
            block_number, context=context
        )
        return block['number'], block['timestamp']

    elif block_timestamp is not None:
        block_number = await async_get_block_of_timestamp(
            block_timestamp, context=context
        )
        block = await block_crud.async_get_block(
            block_number, context=context
        )
        return block['number'], block['timestamp']

    else:
        raise Exception('must specify start_time or block_number')



from __future__ import annotations

from .. import query_utils
from . import contract_creation_blocks_statements


async_query_contract_creation_block = query_utils.wrap_selector_with_connection(
    contract_creation_blocks_statements.async_select_contract_creation_block,
    'contract_creation_blocks',
)


async_query_contract_creation_blocks = query_utils.wrap_selector_with_connection(
    contract_creation_blocks_statements.async_select_contract_creation_blocks,
    'contract_creation_blocks',
)



from __future__ import annotations

import toolcli

from ctc import evm
from ctc import rpc
from ctc import spec
from ctc.protocols import rari_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_fuse_command,
        'help': 'summarize fuse pool, token, or platform',
        'args': [
            # TODO: allow arg to be pool comptroller
            # - will need to check on chain to see if token or comptroller
            {
                'name': 'arg',
                'nargs': '?',
                'help': 'pool index or token, omit to summarize platform',
            },
            {'name': '--block', 'default': 'latest', 'help': 'block number'},
        ],
        'examples': {
            '8': {'description': 'summarize a pool'},
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca': {
                'description': 'summarize a token across all pools',
                'long': True,
            },
            '': {'description': 'summarize all of Rari', 'long': True},
        },
    }


async def async_fuse_command(arg: str, block: str) -> None:
    block_number = await evm.async_block_number_to_int(block)

    if arg is not None:

        try:
            pool_index = int(arg)
            arg_type = 'pool_index'
        except Exception:
            token = arg
            arg_type = 'token'

        if arg_type == 'pool_index':
            await async_print_fuse_pool_summary(
                pool_index=pool_index, block=block_number
            )
        elif arg_type == 'token':
            await async_print_fuse_token_summary(token, block_number)

    else:
        await async_print_all_fuse_pools_summary(block=block_number)


async def async_print_fuse_pool_summary(
    pool_index: int, block: spec.BlockNumberReference
) -> None:
    all_pools = await rari_utils.async_get_all_pools(block=block)
    pool = all_pools[pool_index]
    comptroller = pool[2]
    standard_block = evm.standardize_block_number(block)
    block_data = await rpc.async_eth_get_block_by_number(
        block_number=standard_block,
        include_full_transactions=False,
    )

    tokens_data = await rari_utils.async_get_pool_summary(
        comptroller=comptroller,
        block=block,
    )
    pool_name = await rari_utils.async_get_pool_name(comptroller=comptroller)
    rari_utils.print_fuse_pool_summary(
        block=block_data,
        tokens_data=tokens_data,
        pool_name=pool_name,
        comptroller=comptroller,
    )


async def async_print_all_fuse_pools_summary(
    block: spec.BlockNumberReference,
) -> None:

    await rari_utils.async_print_all_pool_summary(block=block)


async def async_print_fuse_token_summary(
    token: spec.Address,
    block: spec.BlockNumberReference,
) -> None:
    if token == 'ETH':
        token = '0x0000000000000000000000000000000000000000'
    await rari_utils.async_print_fuse_token_summary(token=token, block=block)



from __future__ import annotations

import toolcli

from ctc import evm
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': checksum_command,
        'help': 'compute checksum of address',
        'args': [
            {'name': 'address', 'help': 'address to get checksum of'},
        ],
        'examples': [
            '0xd8da6bf26964af9d7eed9e03e53415d37aa96045',
        ],
    }


def checksum_command(address: spec.Address) -> None:
    print(evm.get_address_checksum(address))



from __future__ import annotations

import ast
import re

import toolcli

from ctc import evm


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': rlp_encode_command,
        'help': 'RLP encode data',
        'args': [
            {
                'name': 'data',
                'help': 'data to encode (can enclose with quotes)',
            },
        ],
        'examples': [
            '1024',
            'cat',
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca',
        ],
    }


def rlp_encode_command(data: str) -> None:

    # quote hex data
    regex = re.compile('0x[0-9a-fA-F]+')
    for match in regex.findall(data):
        data = data.replace(match, '"' + match + '"')

    # convert data to primitive python types
    try:
        parsed = ast.literal_eval(data)
    except ValueError:
        parsed = ast.literal_eval('"""' + data + '"""')

    # rlp encode
    encoded = evm.rlp_encode(parsed)

    # output encoded data
    print(encoded)



from __future__ import annotations

import typing

import msgspec


#
# # msgspec
#

CreateAction = msgspec.defstruct(
    'CreateAction',
    [
        ('from', typing.Optional[str], None),  # type: ignore
        ('init', typing.Optional[str], None),  # type: ignore
    ],
    omit_defaults=True,
)

CreateResult = msgspec.defstruct(
    'CreateResult',
    [
        ('address', typing.Optional[str], None),  # type: ignore
        ('code', typing.Optional[str], None),  # type: ignore
    ],
)


class CreateTrace(msgspec.Struct, omit_defaults=True, rename='camel'):
    type: str
    trace_address: list[int]
    subtraces: int
    action: typing.Optional[CreateAction] = None  # type: ignore
    result: typing.Optional[CreateResult] = None  # type: ignore
    transaction_hash: typing.Optional[str] = None
    error: typing.Optional[str] = None


class RpcResult(msgspec.Struct):
    result: list[CreateTrace]


decoder = msgspec.json.Decoder(RpcResult)


#
# # functions
#


def decode_create_traces(raw_block_trace: str, block_number: int) -> typing.Sequence[CreateTrace]:

    response = decoder.decode(raw_block_trace)
    call_traces = response.result

    from . import native_transfer_decoder
    call_traces = native_transfer_decoder.filter_failed_traces(call_traces)

    create_traces: list[CreateTrace] = []
    for t, trace in enumerate(call_traces):
        if trace.type == 'create' and trace.result is not None:

            # identify root trace for deployer
            d = t
            while call_traces[d].trace_address != []:
                d = d - 1
                if d < 0:
                    raise Exception('could not find deployer of trace')

            # create trace
            create_trace = {
                'block_number': block_number,
                'create_index': len(create_traces),
                'transaction_hash': trace.transaction_hash,
                'contract_address': trace.result.address,
                'deployer': getattr(call_traces[d].action, 'from'),
                'factory': getattr(trace.action, 'from'),
                'init_code': trace.action.init,
                'code': trace.result.code,
            }

            create_traces.append(create_trace)

    return create_traces




from __future__ import annotations

import toolsql

block_gas_stats_schema: toolsql.DBSchemaShorthand = {
    'tables': {
        'block_gas_stats': {
            'columns': [
                {'name': 'block_number', 'type': 'Integer', 'primary': True},
                {'name': 'base', 'type': 'Integer'},
                {'name': 'mean', 'type': 'Integer'},
                {'name': 'median', 'type': 'Integer'},
                {'name': 'min', 'type': 'Integer'},
                {'name': 'max', 'type': 'Integer'},
                {'name': 'p1', 'type': 'Integer'},
                {'name': 'p5', 'type': 'Integer'},
                {'name': 'p10', 'type': 'Integer'},
                {'name': 'p25', 'type': 'Integer'},
                {'name': 'p75', 'type': 'Integer'},
                {'name': 'p90', 'type': 'Integer'},
                {'name': 'p95', 'type': 'Integer'},
                {'name': 'p99', 'type': 'Integer'},
                {'name': 'limit', 'type': 'Integer'},
                {'name': 'used', 'type': 'Integer'},
                {'name': 'n_transactions', 'type': 'Integer'},
            ],
        },
    },
}



from __future__ import annotations

import os
import typing

import toolcli

from ctc import evm

if typing.TYPE_CHECKING:
    from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_abi_command,
        'help': 'display abi of contract',
        'args': [
            {'name': 'address', 'help': 'address of contract'},
            {
                'name': 'names',
                'help': 'name of function or event',
                'nargs': '*',
            },
            # {
            #     'name': '--rw',
            #     'help': 'sort function entries by read vs write functions',
            #     'action': 'store_true',
            #     'dest': 'read_write',
            # },
            {
                'name': '--json',
                'dest': 'json_pretty',
                'help': 'json only',
                'action': 'store_true',
            },
            {
                'name': '--json-raw',
                'help': 'json only, without sorting or indentations',
                'action': 'store_true',
            },
            {
                'name': '--map-names',
                'help': 'output ABI as map where names are the keys',
                'action': 'store_true',
            },
            {
                'name': '--map-selectors',
                'help': 'output ABI as map where hash selectors are the keys',
                'action': 'store_true',
            },
            {
                'name': ['-f', '--functions'],
                'help': 'display function abi\'s only',
                'action': 'store_true',
            },
            {
                'name': ['-e', '--events'],
                'help': 'display event abi\'s only',
                'action': 'store_true',
            },
            {
                'name': '--search',
                'help': 'query of name of function or event abi',
            },
            {
                'name': ['--verbose', '-v'],
                'help': 'include extra abi data',
                'action': 'store_true',
            },
            {
                'name': '--python',
                'help': 'output data as pep8-formatted python syntax',
                'action': 'store_true',
            },
            {
                'name': '--update',
                'help': 're-import ABI from etherscan (e.g. if proxy has changed)',
                'action': 'store_true',
            },
            {
                'name': '--network',
                'help': 'network of context',
            },
        ],
        'examples': [
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca',
            '0x2b79b3c3c7b35463a28a76e0d332aab3e20aa337 Mint Burn Swap Sync',
            '0x2b79b3c3c7b35463a28a76e0d332aab3e20aa337 -f',
            '0x2b79b3c3c7b35463a28a76e0d332aab3e20aa337 -e',
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca --json',
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca --map-names --python',
        ],
    }


async def async_abi_command(
    *,
    address: spec.Address,
    names: typing.Sequence[str],
    # read_write: bool,
    json_pretty: bool,
    json_raw: bool,
    functions: bool,
    events: bool,
    search: str,
    verbose: bool,
    map_names: bool,
    map_selectors: bool,
    python: bool,
    update: bool,
    network: spec.NetworkReference,
) -> None:

    if map_names and map_selectors:
        raise Exception('can only specify one of --map-names or --map-keys')
    if map_names or map_selectors:
        if not json_raw:
            json_pretty = True

    address = await evm.async_resolve_address(address)
    contract_abi = await evm.async_get_contract_abi(
        contract_address=address,
        context={'network': network, 'cache': (not update)},
    )

    # filter by name
    if len(names) > 0:
        contract_abi = [
            item
            for item in contract_abi
            if item.get('name') is not None
            and any(name == item['name'] for name in names)
        ]
    if search is not None:
        search = search.lower()
        contract_abi = [
            item
            for item in contract_abi
            if item.get('name') is not None and search in item['name'].lower()
        ]

    # filter by type
    if functions:
        contract_abi = [
            item for item in contract_abi if item.get('type') == 'function'
        ]
    if events:
        contract_abi = [
            item for item in contract_abi if item.get('type') == 'event'
        ]

    # output abis
    json_only = json_pretty or json_raw
    human_only = not json_only
    if not human_only:
        import json

        if map_names:
            json_data: typing.Any = {}
            for item in contract_abi:
                name = item.get('name')
                if name is None:
                    continue
                if name in json_data:
                    raise Exception('ABI naming conflict: ' + str(name))
                json_data[name] = item
        elif map_selectors:
            json_data = evm.get_contract_abi_by_selectors(contract_abi)
        else:
            json_data = contract_abi

        if json_raw:
            as_str = json.dumps(json_data)
        else:
            as_str = json.dumps(json_data, indent=4, sort_keys=True)

        # convert from json to python syntax
        if python:
            import re

            as_str = re.sub('(?<![\[{,])\n', ',\n', as_str)
            as_str = as_str.replace('"', '\'')
            as_str = as_str.replace(': true,\n', ': True,\n')
            as_str = as_str.replace(': false,\n', ': False,\n')

        print(as_str)
    if not json_only and not human_only:
        print()
        print()
    if not json_only:
        columns = toolcli.get_n_terminal_cols()

        if functions:
            evm.print_contract_abi_functions(
                contract_abi,
                max_width=columns,
                verbose=verbose,
                read_write=True,
            )

        elif events:
            evm.print_contract_abi_events(
                contract_abi,
                max_width=columns,
                verbose=verbose,
            )

        else:
            evm.print_contract_abi(
                contract_abi,
                max_width=columns,
                verbose=verbose,
                read_write=True,
            )




from __future__ import annotations

from ctc import rpc
from ctc import spec

from .. import uniswap_v3_spec


async def async_pool_factory(
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> spec.Address:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'factory', 'pool'
    )
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        context=context,
        block_number=block,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result


async def async_pool_token0(
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> spec.Address:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'token0', 'pool'
    )
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        context=context,
        block_number=block,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result


async def async_pool_token1(
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> spec.Address:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'token1', 'pool'
    )
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        context=context,
        block_number=block,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result


async def async_pool_fee(
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    function_abi = await uniswap_v3_spec.async_get_function_abi('fee', 'pool')
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        context=context,
        block_number=block,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_pool_tick_spacing(
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'tickSpacing', 'pool'
    )
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        context=context,
        block_number=block,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_pool_max_liquidity_per_tick(
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'maxLiquidityPerTick',
        'pool',
    )
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        context=context,
        block_number=block,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result




from __future__ import annotations


class ConfigUpgradeError(Exception):
    pass


class ConfigInvalid(ValueError):
    pass


class ConfigDoesNotExist(FileNotFoundError):
    pass


class ConfigPathNotSet(Exception):
    pass




from __future__ import annotations

import typing

from ctc import evm
from ctc import spec

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    class AddressTransactionCounts(TypedDict):
        blocks: list[int]
        diffs: list[int]
        cummulative: list[int]


@typing.overload
async def async_get_transactions_from_address(
    address: spec.Address,
    output_format: typing.Literal['dataframe'],
    *,
    context: spec.Context = None,
) -> spec.DataFrame:
    ...


@typing.overload
async def async_get_transactions_from_address(
    address: spec.Address,
    output_format: typing.Literal['hashes'],
    *,
    context: spec.Context = None,
) -> typing.Sequence[str]:
    ...


@typing.overload
async def async_get_transactions_from_address(
    address: spec.Address,
    output_format: typing.Literal['full'] = 'full',
    *,
    context: spec.Context = None,
) -> typing.Sequence[spec.DBTransaction]:
    ...


@typing.overload
async def async_get_transactions_from_address(
    address: spec.Address,
    output_format: typing.Literal['full', 'dataframe', 'hashes'],
    *,
    context: spec.Context = None,
) -> typing.Sequence[spec.DBTransaction] | spec.DataFrame | typing.Sequence[
    str
]:
    ...


async def async_get_transactions_from_address(
    address: spec.Address,
    output_format: typing.Literal['full', 'dataframe', 'hashes'] = 'full',
    *,
    context: spec.Context = None,
) -> typing.Sequence[spec.DBTransaction] | spec.DataFrame | typing.Sequence[
    str
]:
    """get all transactions from an address"""

    address = address.lower()

    count_data = await async_get_address_transaction_counts_by_block(
        address, context=context
    )

    # note: this can be made much more efficient with proper queries
    transactions = []
    for block_number, block_count in zip(
        count_data['blocks'],
        count_data['diffs'],
    ):
        block_transactions = await evm.async_get_block_transactions(
            block=block_number,
            context=context,
        )
        n_block_transactions = 0
        for transaction in block_transactions:
            if transaction['from_address'] == address:
                transactions.append(transaction)
                n_block_transactions += 1
                if n_block_transactions == block_count:
                    break
    if output_format == 'full':
        return transactions
    elif output_format == 'dataframe':
        import polars as pl

        return pl.DataFrame(transactions)
    elif output_format == 'hashes':
        return [transaction['hash'] for transaction in transactions]
    else:
        raise Exception('unknown output format: ' + str(output_format))


async def async_get_address_transaction_counts_by_block(
    address: spec.Address,
    nary: int = 3,
    *,
    context: spec.Context = None,
) -> AddressTransactionCounts:
    """return historical transaction count of address by block"""

    import asyncio
    from ctc import rpc

    address = address.lower()

    # get initial data
    min_block = 0
    min_count_coroutine = rpc.async_eth_get_transaction_count(
        from_address=address,
        block_number=0,
        context=context,
    )
    min_count_task = asyncio.create_task(min_count_coroutine)
    max_block = await rpc.async_eth_block_number(context=context)
    max_count = await rpc.async_eth_get_transaction_count(
        from_address=address,
        block_number=max_block,
        context=context,
    )
    min_count = await min_count_task

    block_counts = {
        min_block: min_count,
        max_block: max_count,
    }

    await _async_get_block_range_transaction_counts(
        address=address,
        min_block=min_block,
        max_block=max_block,
        block_counts=block_counts,
        nary=nary,
        context=context,
    )

    # parse blocks that contain transactions
    blocks = []
    diffs = []
    cummulative = []
    as_tuples = sorted(block_counts.items())
    for before, after in zip(as_tuples[:-1], as_tuples[1:]):
        if before[0] + 1 == after[0] and before[1] != after[1]:
            blocks.append(after[0])
            diffs.append(after[1] - before[1])
            cummulative.append(after[1])

    return {
        'blocks': blocks,
        'diffs': diffs,
        'cummulative': cummulative,
    }


async def _async_get_block_range_transaction_counts(
    *,
    address: spec.Address,
    min_block: int,
    max_block: int,
    block_counts: dict[int, int],
    nary: int,
    context: spec.Context = None,
) -> None:

    import asyncio
    import numpy as np
    from ctc import rpc

    n_unknown_blocks = max_block - min_block - 1
    if n_unknown_blocks <= 0:
        return

    blocks = np.linspace(
        min_block,
        max_block,
        min(nary + 1, n_unknown_blocks + 2),
    )[1:-1].astype(int)

    coroutines = [
        rpc.async_eth_get_transaction_count(
            from_address=address,
            block_number=block,
            context=context,
        )
        for block in blocks
    ]
    counts = await asyncio.gather(*coroutines)
    for block, count in zip(blocks, counts):
        block_counts[block] = count

    all_blocks = [min_block] + list(blocks) + [max_block]

    recursive_coroutines = []
    for range_min_block, range_max_block in zip(
        all_blocks[:-1], all_blocks[1:]
    ):
        range_min_count = block_counts[range_min_block]
        range_max_count = block_counts[range_max_block]

        if range_min_count == range_max_count:
            continue
        elif (
            range_min_count + 1 == range_max_count
            and range_min_block + 1 == range_max_block + 1
        ):
            continue
        else:
            coroutine = _async_get_block_range_transaction_counts(
                address=address,
                min_block=range_min_block,
                max_block=range_max_block,
                block_counts=block_counts,
                nary=nary,
                context=context,
            )
            recursive_coroutines.append(coroutine)
    await asyncio.gather(*recursive_coroutines)




from __future__ import annotations

from ... import query_utils
from . import transactions_statements


async_query_transaction = query_utils.wrap_selector_with_connection(
    transactions_statements.async_select_transaction,
    'transactions',
)

async_query_transactions = query_utils.wrap_selector_with_connection(
    transactions_statements.async_select_transactions,
    'transactions',
)

async_query_block_transaction_query = query_utils.wrap_selector_with_connection(
    transactions_statements.async_select_block_transaction_query,
    'transactions',
)

async_query_block_transaction_queries = (
    query_utils.wrap_selector_with_connection(
        transactions_statements.async_select_block_transaction_queries,
        'transactions',
    )
)

async_query_block_transactions = query_utils.wrap_selector_with_connection(
    transactions_statements.async_select_block_transactions,
    'transactions',
)

async_query_blocks_transactions = query_utils.wrap_selector_with_connection(
    transactions_statements.async_select_blocks_transactions,
    'transactions',
)




from __future__ import annotations

import toolcli
import toolstr

import ctc.config
from ctc import evm
from ctc import rpc
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_chain_command,
        'help': 'display chain_id of provider',
        'args': [
            {'name': 'provider', 'help': 'url or configured name of RPC node'},
            {
                'name': ['--verbose', '-v'],
                'help': 'display additional information about provider chain',
                'action': 'store_true',
            },
            {
                'name': '--config',
                'help': 'display only chain_id from config, not reported chain_id',
                'action': 'store_true',
            },
            {
                'name': '--no-check',
                'help': 'do not check reported chain_id against configured chain_id',
                'action': 'store_true',
            },
        ],
        'examples': [
            'https://rpc.flashbots.net/',
            'https://arbitrumrpc.com',
        ],
    }


async def async_chain_command(
    *,
    provider: str,
    config: bool,
    verbose: bool,
    no_check: bool,
) -> None:

    context = ctc.config.create_user_input_context(provider=provider)

    # gather configured chain_id
    config_chain_id = None
    if verbose or config or not no_check:
        try:
            config_chain_id = ctc.config.get_context_chain_id(context)
        except spec.CouldNotDetermineNetwork:
            pass

    # gather reported chain_id
    reported_chain_id = None
    if verbose or not config or not no_check:
        reported_chain_id = await rpc.async_eth_chain_id(context=context)

    # print output
    if verbose:
        toolstr.print_text_box('Provider chain information')
        print()

        reported_chain_id = await rpc.async_eth_chain_id(context=context)
        toolstr.print_header('Info reported by node')
        print('- chain_id:', reported_chain_id)
        print('- name:', evm.get_network_name(reported_chain_id))
        print()
        toolstr.print_header('Info in config')
        if config_chain_id is None:
            print('[provider not in config]')
        else:
            print('- chain_id:', config_chain_id)
            print('- name:', evm.get_network_name(config_chain_id))

    else:
        if config:
            print(config_chain_id)
        else:
            print(reported_chain_id)

    # perform check
    if not no_check:
        if (
            reported_chain_id is not None
            and config_chain_id is not None
            and reported_chain_id != config_chain_id
        ):
            print()
            raise Exception(
                '[WARNING] configured chain_id is '
                + str(config_chain_id)
                + ' but chain_id reported by node is '
                + str(reported_chain_id)
            )
